-- $Id: object_flag_callbacks.lua 3564 2019-01-22 04:20:44Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2019 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

-- Actor callbacks to be added to objects in Object:resolve() based on the
-- presence of various object flags.

newFlagCallback {
  group = 'flags',
  flag = 'AUTO_CURSE',
  worn_callbacks = {
    on_global_turn = function(who, o)
      if rng.chance(15) then o:setCursed(true) end
    end,
  },
}

newFlagCallback {
  group = 'flags',
  flag = 'TY_CURSE',
  worn_callbacks = {
    on_global_turn = function(who, o)
      if rng.chance(100) then obj_util.apply_ty_curse(o, who) end
    end,
  },
}

newFlagCallback {
  group = 'flags',
  flag = 'DG_CURSE',
  worn_callbacks = {
    on_global_turn = function(who, o)
      if rng.chance(50) then obj_util.apply_dg_curse(o, who) end
    end,
  },
}

newFlagCallback {
  group = 'convey_flags',
  flag = 'TELEPORT',
  worn_callbacks = {
    on_global_turn = function(who, o)
      if rng.chance(100) then obj_util.obj_teleport(o, who) end
    end,
  },
}

newFlagCallback {
  group = 'flags',
  flag = 'CLONE',
  worn_callbacks = {
    on_melee_blow = function(who, o, target, hit, wpn)
      if hit and wpn == o and rng.percent(30) then
	obj_util.obj_clone(target)
      end
    end
  },
}

newFlagCallback {
  group = 'flags',
  flag = 'VORPAL',
  worn_callbacks = {
    -- Vorpal effects (cf. src/cmd1.c lines 2788-2799
    on_wpn_hit_dam_mod = function(who, o, target, wpn, p)
      if wpn == o and rng.chance(6) then
	local rep = p.base
	game.logPlayer(who, 'Your weapon cuts deep into %s!', target:getName{def_art=true, check_seen=true})
	repeat p.base = p.base + rep until not rng.chance(4)
      end
    end
  },
}

newFlagCallback {
  group = 'flags',
  flag = 'WOUNDING',
  worn_callbacks = {
    -- Special cut effect (cf. src/cmd1.c lines 495-511)
    on_wpn_hit_pre = function(who, o, target, wpn, p)
      if wpn == o and not p.stop and not target.dead then
	if target:has('NO_CUT') and target.learnFlags then
	  target:learnFlags('flags', 'NO_CUT')
	elseif rng.percent(50) and not rng.percent(target.level) then
	  target:setEffect(target.EFF_CUT, base*2, {merge='combine', src=who, combine_msg='@Target@ is bleeding more strongly.'})
	end
      end
    end
  },
}

newFlagCallback {
  group = 'flags',
  flag = 'CHAOTIC',
  worn_callbacks = {
    -- Chaotic effects (cf. src/cmd1.c lines 2688-2715)
    on_wpn_hit_pre = function(who, o, target, wpn, p)
      if wpn == o and rng.chance(2) and not p.stop and not target.dead then
	-- Note:  The probabilities computed below don't match the comments
	-- in this part of the T2 code, but they do match the actual code.
	-- We follow the latter.
	if rng.chance(40) then
	  -- Vampiric effects (prob. 40%)
	  -- Spectacularly Dodgy Hack(TM):  We have to decide to do this
	  -- now, but we can't do it until after we know how much weapon
	  -- damage we're going to do.
	  obj_util.once_per_wpn_hit.delayed_vampiric_drain = true
	elseif rng.chance(250) then
	  -- Earthquake (prob. 0.24%)
	  -- Spectacularly Dodgy Hack(TM), continued:  Weapon-triggered
	  -- earthquakes happen after all melee attacks have been processed.
	  obj_util.once_per_melee_attack.delayed_earthquake = true
	elseif rng.chance(90) then
	  -- Confusion (prob. 53.784%)
	  obj_util.wpn_confuse(who, target)
	elseif rng.chance(2) then
	  -- Teleport away (prob. 2.988%)
	  local t_name = target:getName{def_art=true, check_seen=true}
	  game.logSeen(target, '%s disappears!', t_name:capitalize())
	  spell.teleport_away(who, target, 50)
	  p.stop = true
	else
	  -- Polymorph (prob. 2.988%)
	  -- TODO
	end
      end
    end,
    on_wpn_hit_dam_mod = function(who, o, target, wpn, p)
      if wpn == o and obj_util.once_per_wpn_hit.delayed_vampiric_drain then
	-- Spectacularly Dodgy Hack(TM), redux:  We triggered vampiric
	-- drain in the earlier callback; now that we know how much damage
	-- we're about to do, we can do it.
	obj_util.once_per_wpn_hit.delayed_vampiric_drain = nil
	obj_util.wpn_vampric_drain(who, target, p.base + p.bonus)
      end
    end,
    on_melee_attack = function(who, o, _, _, _)
      if obj_util.once_per_melee_attack.delayed_earthquake then
	-- Spectacularly Dodgy Hack(TM), redux and epilogue:  We chose
	-- earthquake in the earlier callbacks; trigger it now.
	obj_util.once_per_melee_attack.delayed_earthquake = nil
	spell.earthquake(who.x, who.y, 10)
      end
    end,
  },
}

newFlagCallback {
  group = 'flags',
  flag = 'VAMPIRIC',
  worn_callbacks = {
    on_wpn_hit_dam_mod = function(who, o, target, wpn, p)
      if wpn == o and not p.stop and not target.dead then
	obj_util.wpn_vampric_drain(who, target, p.base + p.bonus)
      end
    end
  },
}

newFlagCallback {
  group = 'convey_flags',
  flag = 'IMPACT',
  worn_callbacks = {
    on_wpn_hit_dam_mod = function(who, o, target, wpn, p)
      if wpn == o and (p.base > 50 or rng.chance(7)) and not target.dead then
	-- Spectacularly Dodgy Hack(TM):  Weapon-triggered earthquakes
	-- happen after all melee attacks have been processed.
	obj_util.once_per_melee_attack.delayed_earthquake = true
      end
    end,
    on_melee_attack = function(who, o, _, _, _)
      if obj_util.once_per_melee_attack.delayed_earthquake then
	-- Spectacularly Dodgy Hack(TM), redux:  Trigger the earthquake we
	-- set up in the previous callback.
	obj_util.once_per_melee_attack.delayed_earthquake = nil
	spell.earthquake(who.x, who.y, 10)
      end
    end,
  },
}
