-- $Id: lothlorien-poisoned-water.lua 3601 2019-05-11 17:56:57Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2018-2019 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

add_hooks = function()
  class:bindHook('CharacterSheet:dumpLine', function(self, data)
    local q = game.player:hasQuest('lothlorien-poisoned-water')
    if q and q:isEnded() then
      table.insert(data.lines, [[You saved the beautiful Mallorns of Lothlórien.]])
    end
  end)
end

id = 'lothlorien-poisoned-water'
name = 'Poisoned Water'
danger_level = 30	-- cf. src/tables.c line 3896
reward = 'something'	-- populated by on_status_change() below

desc = function(self, who)
  if self:isEnded() then
    return ([[You cleared out the mold infestation that was poisoning the river Nimrodel and purified its waters.  As a reward, Celeborn gave you %s.]]):format(self.reward)
  elseif self:isCompleted() then
    return [[You cleared out the mold infestation that was poisoning the river Nimrodel and purified its waters.  You should return to the Seat of Ruling in Lothlórien and tell Celeborn.]]
  elseif self:isCompleted('kill-mold') then
    return [[You have cleared out the mold infestation that was poisoning the river Nimrodel.  Their poison still lingers in the water, though; you should drop your Potions of Water Curing into the tainted water to purify it.]]
  else
    return [[You must search the waters of the river Nimrodel west of Lothlórien to find the source of the poison that is tainting its waters.  Celeborn has given you Potions of Water Curing to purify the waters after you have removed the source of the poison.]]
  end
end

on_grant = function(self, who)
  local lxy = game.state:getWorldMapZoneCoords('town-lothlorien')
  local cand_locs = {
    { x = lxy.x - 2, y = lxy.y - 2 },
    { x = lxy.x - 1, y = lxy.y - 2 },
    { x = lxy.x - 2, y = lxy.y - 1 },
    { x = lxy.x - 2, y = lxy.y },
  }
  local loc = rng.table(cand_locs)
  -- We may not need this here, but just in case...
  self.map_loc = loc
  game.state:setWildernessQuest(loc.x, loc.y, self.id)

  local o = game.zone:makeEntityByName(game.level, 'object', 'POT_WATER_CURING')
  if o then
    o:setNumber(99)
    o:identify(true, who)
    o:setFoundHere {
      type = 'custom',
      custom = 'Celeborn gave %s to you to purify the waters of Nimrodel.',
    }
    game.logPlayer(who, 'Celeborn gives you:  %s', o:getName())
    who:addOrDropObject(who.INVEN_INVEN, o, nil, nil, {vocal_drop=true})
    who:sortInven()
  end
end

on_status_change = function(self, who, status, sub)
  if status == self.DONE then
    local Object = require 'mod.class.Object'
    Object.resolve_flags.force_ego = 'ELVEN'
    local o = game.zone:makeEntityByName(game.level, 'object', 'BLUE_DRAGON_SCALE_MAIL')
    if o then
      o:identify(true, who)
      o:setFoundHere {
	type = 'custom',
	custom = 'You received %s as a reward for purifying the poisoned waters of Nimrodel.',
      }
      self.reward = o:getName()
      game.logPlayer(who, 'Celeborn gives you:  %s', o:getName())
      who:addOrDropObject(who.INVEN_INVEN, o, nil, nil, {vocal_drop=true})
      who:sortInven()
    end
    world:gainAchievement('POISON_WATER', who)
  elseif status == self.COMPLETED and not sub then
    -- Clear the tainted water.
    game:onTickEnd(function ()
      local Map = require 'mod.class.Map'
      local map = game.level.map
      for x = 0, map.w - 1 do
	for y = 0, map.h - 1 do
	  local g = map(x, y, Map.TERRAIN)
	  if g.define_as == 'TAINTED_WATER' then
	    map(x, y, Map.TERRAIN, game.zone.grid_list.SHALLOW_WATER)
	  end
	end
      end
    end)
    local Dialog = require 'engine.ui.Dialog'
    Dialog:simpleLongPopup(self.name, [[As you drop the potions into the river, they shatter and spread out into the waters.  Almost immediately, the taint of the poison begins to clear.]], 600)
  elseif status == self.COMPLETED and sub == 'kill-mold' then
    local Dialog = require 'engine.ui.Dialog'
    Dialog:simpleLongPopup(self.name, [[You have killed enough of the molds to weaken their poison.  If you drop your Potions of Water Curing into the water now, they should be effective.]], 600)
  end
end

on_quest_mon_die = function(self, who, src)
  -- Only interested in kills in the quest wilderness zone.
  if not game.zone.wilderness then return end
  local q_id, _ = game.state:getWildernessQuest()
  if q_id ~= self.id then return end
  -- Unlike most quests of this nature, we don't need to kill all
  -- hostiles, just most of them.
  local n_mold = 0
  for _, e in pairs(game.level.entities) do
    if e.reactionToward and e:reactionToward(game.player) < 0 and not e.dead then
      n_mold = n_mold + 1
    end
  end
  if n_mold < 10 and not self:isCompleted('kill-mold') then
    -- :on_status_change() will pop up a dialog for us.
    game.player:setQuestStatus(self.id, self.COMPLETED, 'kill-mold')
  end
end

callbacks = {
  on_drop_pre = function(who, q, o, silent, p)
    -- Don't care about anything but the Potions of Water Curing.
    if o.define_as ~= 'POT_WATER_CURING' then return end
    -- cf. quest_poison_drop_hook() in src/q_poison.c:
    -- Only if we're in the quest wilderness zone.
    local ok_loc = game.zone.wilderness
    if ok_loc then
      local q_id, _ = game.state:getWildernessQuest()
      ok_loc = ok_loc and q_id == 'lothlorien-poisoned-water'
    end
    if not ok_loc then
      game.log('You should hang onto these until you find the poisoned water in Nimrodel.')
      p.no_drop = true
      return
    end
    -- Only if we've killed enough of the molds.
    local n_mold = 0
    for _, e in pairs(game.level.entities) do
      if e.reactionToward and e:reactionToward(who) <= 0 and not e.dead then
	n_mold = n_mold + 1
      end
    end
    if n_mold >= 10 then
      game.log('The source of the poison is still too strong.  You need to clear out more of the mold infestation.')
      p.no_drop = true
      return
    end
    -- Sneaky Hack(TM):  Pull the Potions out of inventory manually and
    -- discard them.
    local _, item, inven_id = who:findInAllInventoriesBy('define_as', 'POT_WATER_CURING')
    if item and inven_id then
      local oo = who:removeObject(inven_id, item, true)
      if game:hasEntity(oo) then game:removeEntity(oo) end
      -- on_status_change() will clear the waters for us.
      game.player:setQuestStatus(q.id, q.COMPLETED)
      -- Still need to tell caller not to drop them, since they're no
      -- longer there to be dropped. ;)
      p.no_drop = true
      -- Annoyingly, the preceding will prevent the inventory dialog from
      -- closing; we close them manually.
      local to_close = {}
      for _, d in ipairs(game.dialogs) do
	if d.inventoryLiveUpdate then to_close[#to_close+1] = d end
      end
      for _, d in ipairs(to_close) do game:unregisterDialog(d) end
      return
    end
  end,
}

-- Hooks into the wilderness-quest infrastructure:

-- Don't generate any actors other than the ones we place manually.
merge_actor_gen = {
  no_prepop = true,
}

-- cf. quest_poison_gen_hook() in src/q_poison.c
post_process_level = function(self, level, zone)
  -- Pick a location for our blob of molds.
  local Map = require 'mod.class.Map'
  local map = game.level.map
  local cx, cy, tries = 1, 1
  for _ = 1, 10000 do
    local x = rng.range(1, map.w - 34) + 32
    local y = rng.range(1, map.h - 24) + 22
    if map:checkEntity(x, y, Map.TERRAIN, 'empty_bold') then
      cx, cy = x, y
      break
    end
  end
  -- Taint water in a radius of 25, and place molds in a radius of 10.
  local gen = game.zone:getGenerator('actor', game.level)
  -- cf. create_molds_hook() in src/q_poison.c
  local filter = {
    displays = 'm,e',
    ignore = {
      deep_flags = { flags={MULTIPLY=true} },
    }
  }
  local function no_block(_, x, y) return false end
  local function cb(_, x, y, dx, dy, sqdist)
    local g = map(x, y, Map.TERRAIN)
    if rng.percent(80) and (g.define_as == 'DEEP_WATER' or g.define_as == 'SHALLOW_WATER') then
      map(x, y, Map.TERRAIN, game.zone.grid_list.TAINTED_WATER)
    end
    if sqdist <= 10*10 and rng.percent(60) then
      local m = gen:generateOne{ filter=filter, x=x, y=y, no_grp=true, ret=true }
      if m then
	m.quest_on_die = self.id
	-- Some of these we level up.
	if rng.percent(80) and m.level < game.player.level then
	  m:forceLevelup(rng.range(m.level+1, game.player.level))
	end
      end
    end
  end
  core.fov.calc_circle(cx, cy, map.w, map.h, 25, no_block, cb, nil)
end

worldmap_change_grid = function(self, x, y, orig_grid)
  if self:isStatus(self.PENDING) then
    local Map = require 'mod.class.Map'
    local g = orig_grid:cloneFull()
    g.name = 'a tainted water area'
    g.display = '~'
    g.color_r = 0x80
    g.color_g = 0x40
    g.color_b = 0x00
    g.image = 'terrain/tainted_water.png'
    return g
  else
    return nil
  end
end
