-- $Id: monster_pit.lua 3161 2017-11-10 02:04:51Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2016-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

-- Roughly mimics build_type6() in src/generate.c, to build a monster pit.
newRoom {
  id = 'monster_pit',
  min_depth = 5,		-- cf. roomdep[] in src/generate.c
  wgt_very_unusual = 15,	-- cf. src/generate.c lines 6942-6943
  crowded = true,
  filters = {
    orc = {
      -- cf. vault_aux_orc() in src/generate.c
      precog_name = 'orc',
      add_levels = 10,
      not_properties = { 'unique' },
      display = 'o',
    },
    troll = {
      -- cf. vault_aux_troll() in src/generate.c
      precog_name = 'troll',
      add_levels = 10,
      not_properties = { 'unique' },
      display = 'T',
    },
    giant = {
      -- cf. vault_aux_giant() in src/generate.c
      precog_name = 'giant',
      add_levels = 10,
      not_properties = { 'unique' },
      display = 'P',
    },
    chapel = {
      -- cf. vault_aux_chapel() in src/generate.c
      precog_name = 'ordered chapel',
      add_levels = 10,
      not_properties = { 'unique' },
      special = function(e)
	return e.display == 'A' or e.name:find('riest', 1, true)
      end,
    },
    demon = {
      -- cf. vault_aux_demon() in src/generate.c
      precog_name = 'demon',
      add_levels = 10,
      not_properties = { 'unique' },
      display = 'U',
    },
  },
  build = function(self, gen, gx, gy)
    local ctr = gen:roomAlloc(25, 11, true, gx, gy)
    if not ctr then return false end

    -- Outer room extents
    local x1, x2, y1, y2 = ctr.x - 11, ctr.x + 11, ctr.y - 4, ctr.y + 4
    -- Floor
    gen:build_floor(x1, y1, x2, y2, 'floor', false, {'room'})
    -- Walls
    gen:build_rectangle(x1-1, y1-1, x2+1, y2+1, 'outer_wall', false, {'room'})

    -- Inner room extents.
    x1, y1, x2, y2 = x1+2, y1+2, x2-2, y2-2
    -- Walls
    gen:build_rectangle(x1-1, y1-1, x2+1, y2+1, 'inner_wall', false, {'room'})
    -- Secret door on one wall.
    gen:place_rect_secret_door(x1-1, y1-1, x2+1, y2+1, ctr.x, ctr.y)

    -- Choose a pit type.
    local tmp = rng.range(1, gen.dun_lev)
    local filter = nil
    if tmp <= 20 then
      filter = self.filters.orc
    elseif tmp <= 40 then
      filter = self.filters.troll
    elseif tmp <= 55 then
      filter = self.filters.giant
    elseif tmp <= 70 then
      if rng.chance(4) then
	filter = self.filters.chapel
      else
	-- Pick a monster to symbol-clone
	-- (cf. src/generate.c lines 3277-3289)
	local tmpl
	repeat
	  tmpl = rng.table(gen.zone.npc_list)
	until not tmpl.unique and not tmpl.special_gene and tmpl.level + rng.range(1, 5) <= gen.dun_lev + rng.range(1, 5)
	filter = {
	  precog_name = 'ordered clones',
	  add_levels = 10,
	  not_properties = { 'unique' },
	  display = tmpl.display,
	}
      end
    elseif tmp <= 80 then
      -- Loosely mimics vault_aux_dragon() in src/generate.c, along with
      -- src/generate.c lines 3302-3388.
      local Talents = require 'engine.interface.ActorTalents'
      local breaths = {
	Talents.NPC_BR_ACID, Talents.NPC_BR_ELEC, Talents.NPC_BR_FIRE,
	Talents.NPC_BR_COLD, Talents.NPC_BR_POIS,
      }
      local dragon_types = {
	{ 'acid dragon',	{ Talents.NPC_BR_ACID } },
	{ 'electric dragon',	{ Talents.NPC_BR_ELEC } },
	{ 'fire dragon',	{ Talents.NPC_BR_FIRE } },
	{ 'cold dragon',	{ Talents.NPC_BR_COLD } },
	{ 'poison dragon',	{ Talents.NPC_BR_POIS } },
	{ 'multi-hued dragon',	breaths },
      }
      local dragon_type = rng.table(dragon_types)
      filter = {
	precog_name = dragon_type[1],
	add_levels = 10,
	not_properties = { 'unique', 'special_gene' },
	displays = 'dD',
	talents = dragon_type[2],
      }
    else
      filter = self.filters.demon
    end

    -- Pick some monsters with our filter.
    local what, empty = {}, false
    for i = 1, 16 do
      what[i] = gen.zone:chooseEntity(gen.level, 'actor', filter, nil, true)
      empty = empty or not what[i]
    end

    -- Pursuant to the T2 code, we only fill the room if we were able to
    -- fill all 16 slots.
    if empty then return false end

    table.sort(what, function(a, b) return a.level < b.level end)
    -- For no clear reason, after picking 16 monsters, we now discard every
    -- other one.
    for i = 16, 2, -2 do table.remove(what, i) end

    gen:report_precog('Monster nest (%s)', filter and filter.precog_name or '???')

    -- Increase level rating.
    gen.level.rating = (gen.level.rating or 0) + 10
    -- Also, depending on depth, possibly flag the level as "special".
    if gen.dun_lev <= 40 and rng.range(1, gen.dun_lev*gen.dun_lev + 50) < 300 then
      gen.level.special_feeling = true
    end

    -- Now actually place the monsters in the room
    local agen = gen.actor_gen
    -- Top and bottom rows
    for x = ctr.x - 9, ctr.x + 9 do
      agen:generateOne{ name = what[1].define_as, x=x, y=ctr.y-2, no_grp=true }
      agen:generateOne{ name = what[1].define_as, x=x, y=ctr.y+2, no_grp=true }
    end
    -- Middle columns
    for y = ctr.y - 1, ctr.y + 1 do
      for dx = 2, 9 do
	local which = what[math.floor((12-dx)/2)]
	agen:generateOne{ name = which.define_as, x=ctr.x-dx, y=y, no_grp=true }
	agen:generateOne{ name = which.define_as, x=ctr.x+dx, y=y, no_grp=true }
      end
    end
    -- Above/below center monster
    for x = ctr.x - 1, ctr.x + 1 do
      agen:generateOne{ name = what[6].define_as, x=x, y=ctr.y-1, no_grp=true }
      agen:generateOne{ name = what[6].define_as, x=x, y=ctr.y+1, no_grp=true }
    end
    -- Next to center monster
    agen:generateOne{ name = what[7].define_as, x=ctr.x-1, y=ctr.y, no_grp=true }
    agen:generateOne{ name = what[7].define_as, x=ctr.x+1, y=ctr.y, no_grp=true }
    -- Center monster
    agen:generateOne{ name = what[8].define_as, x=ctr.x, y=ctr.y, no_grp=true }
    return true
  end,
}
