-- $Id: combat.lua 3566 2019-01-26 20:15:15Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

local Entity = require 'engine.Entity'

newTalentType {
  name = 'Combat skill powers',
  type = 'skill/combat',
  description = 'Talents conferred by skills in the Combat skill tree',
}

newTalent {
  name = [[Change Melee Style]],
  short_name = 'MELEE_STYLE',
  display_entity = Entity.new { image='talents/change_melee_style.png' },
  type = 'skill/combat',
  mode = 'activated',
  points = 0,
  on_pre_use = function(who, t, silent, dry_run)
    if not who.skills then return false end
    local PlayerCombat = require 'mod.class.interface.PlayerCombat'
    local n = 0
    for _, def in ipairs(PlayerCombat.melee_styles) do
      if def:can_use(who) then n = n + 1 end
    end
    return n > 0
  end,
  action = function(self, t)
    self:selectMeleeStyle(true)
  end,
  info = [[Change the player's melee style.  All characters can use the Weapon Combat style; some skills confer access to additional styles.]],
}

newTalent {
  name = [[Piercing Shots]],
  short_name = 'PIERCING_SHOTS',
  display_entity = Entity.new { image='talents/piercing_shots.png' },
  sust_display = function(p) return 'Prc. Shots' end,
  type = 'skill/combat',
  mode = 'sustained',
  no_energy = true,
  info = function(who, t)
    local n_pierce = math.max(0, math.floor(who:getSkillLevel('COMBAT')/10) - 1)
    local prob = 45 + who:getSkillLevel('ARCHERY')
    local fmt = [[While active, your fired ammo has a %d%% chance of piercing through its target and continuing to more targets behind it (piercing chance increases with Archery skill).  A single ammo can pierce through up to %d target%s in this manner (increases with Combat skill).]]
    return fmt:format(prob, n_pierce, (n_pierce == 1 and '' or 's'))
  end,
  sustain_describe = function(who, t, p)
    local n_pierce = math.max(0, math.floor(who:getSkillLevel('COMBAT')/10) - 1)
    local prob = 45 + who:getSkillLevel('ARCHERY')
    local fmt = [[Your fired ammo has a %d%% chance per target of piercing through up to %d target%s.]]
    return tstring{ fmt:format(prob, n_pierce, (n_pierce == 1 and '' or 's')) }
  end,
  activate = function(who, def) return {} end,
  deactivate = function(who, def) return true end,
}

local function can_forge_ammo(self, t, silent, dry_run)
  if self:has('CONFUSED') then
    if not silent then game.logPlayer(self, 'You are too confused!') end
    return false
  end
  if self:has('BLIND') then
    if not silent then game.logPlayer(self, 'You are blind!') end
    return false
  end
  return true
end

-- cf. do_cmd_archer() in src/cmd7.c
local function create_ammo(self, t)
  local cands_by_type = {
    shots	= { 'ROUNDED_PEBBLE', 'IRON_SHOT', 'MITHRIL_SHOT' },
    arrows	= { 'ARROW', 'SEEKER_ARROW' },
    bolts	= { 'ARROW', 'SEEKER_ARROW' },
  }
  local lev = game.zone:level_adjust_level(game.level, 'object')
  local cands = cands_by_type[t.ammo_type]
  local ammo_type = cands[obj_util.m_bonus(#cands - 1, lev) + 1]
  local Object = require 'mod.class.Object'
  local RF = Object.resolve_flags
  RF.force_good, RF.force_great = true, rng.percent(20)
  local o = game.zone:makeEntityByName(game.level, 'object', ammo_type)
  if not o then
    game.log(game.flash.BAD, '??? no ammo created')
    return nil
  end
  -- The T2 code seems to intend to force randart ammo to one, but I've
  -- never seen that happen in practice.
  o:setNumber(rng.range(15, 30))
  o:identify('full', self)
  o:setFoundHere{ type='ammo_forge' }
  o:setDiscount(90)
  game.logPlayer(self, 'You create:  %s', o:getName{do_color=true})
  return o
end

local function can_create_ammo_from(o)
  return (o.tval == 11 or o.tval == 1) and not o.unique
end

local function create_ammo_from(self, t, junk, inven, item, floor)
  if not junk then return nil end
  local ammo = create_ammo(self, t)
  if not ammo then return nil end
  self:getOneObject(inven, item, floor, {vocal_left=true})
  self:addOrDropObject(self.INVEN_INVEN, ammo, nil, nil, {vocal_add=true, vocal_drop=true})
  return true
end

local function create_ammo_from_junk(self, t)
  local function sel_cb(o, item, floor)
    self:talentDialogReturn(true, {o=o, item=item, floor=floor})
    return false
  end
  local ShowInvenFloor = require('mod.dialogs.ShowInvenFloor')
  local inven = self:getInven('INVEN')
  local ret, p = self:talentDialog(ShowInvenFloor.new(t.name, inven, t.use_item.filter, sel_cb, self))
  if not ret then return nil end
  return t.use_item.action(self, t, p.o, inven, p.item, p.floor)
end

newTalent {
  name = [[Forge Shots]],
  short_name = 'FORGE_SHOTS',
  display_entity = Entity.new { image='talents/forge_shots.png' },
  type = 'skill/combat',
  mode = 'activated',
  points = 0,
  on_pre_use = can_forge_ammo,
  action = function(self, t)
    local function cb(x, y)
      local g = game.level.map(x, y, Map.TERRAIN)
      return g.define_as == 'RUBBLE'
    end
    local tx, ty = util.select_adjacent(self, cb, 'There is no rubbble nearby.', 'There is no rubble there.')
    if not tx or not ty then return nil end
    local ammo = create_ammo(self, t)
    if not ammo then
      game.log(game.flash.BAD, '??? no ammo created')
      return nil
    end
    local g = game.level.map(tx, ty, Map.TERRAIN)
    game.logPlayer(self, g.dig.finish)
    -- TODO Choose suitable terrain for the zone, or remember what terrain
    -- was under the rubble.
    game.level.map(tx, ty, Map.TERRAIN, game.zone.grid_list.FLOOR)
    util.place_obj_under_rubble(g, tx, ty)
    self:addOrDropObject(self.INVEN_INVEN, ammo, nil, nil, {vocal_add=true, vocal_drop=true})
    return true
  end,
  ammo_type = 'shots',
  info = [[Forge sling shots from an adjacent pile of rubble.]],
}

newTalent {
  name = [[Forge Arrows]],
  short_name = 'FORGE_ARROWS',
  display_entity = Entity.new { image='talents/forge_arrows.png' },
  type = 'skill/combat',
  mode = 'activated',
  points = 0,
  on_pre_use = can_forge_ammo,
  use_item = {
    name = 'Forge arrows',
    close_before = true,
    filter = can_create_ammo_from,
    action = create_ammo_from,
  },
  action = create_ammo_from_junk,
  ammo_type = 'arrows',
  info = [[Forge arrows from certain types of junk in your inventory or on the floor.]],
}

newTalent {
  name = [[Forge Bolts]],
  short_name = 'FORGE_BOLTS',
  display_entity = Entity.new { image='talents/forge_bolts.png' },
  type = 'skill/combat',
  mode = 'activated',
  points = 0,
  on_pre_use = can_forge_ammo,
  use_item = {
    name = 'Forge crossbow bolts',
    close_before = true,
    filter = can_create_ammo_from,
    action = create_ammo_from,
  },
  action = create_ammo_from_junk,
  ammo_type = 'bolts',
  info = [[Forge crossbow bolts from certain types of junk in your inventory or on the floor.]],
}

newTalent {
  name = [[Create Boulders]],
  short_name = 'CREATE_BOULDERS',
  display_entity = Entity.new { image='talents/create_boulders.png' },
  type = 'skill/combat',
  mode = 'activated',
  points = 0,
  action = function(self, t)
    game.log('Sorry, not implemented (CREATE_BOULDERS)')
  end,
  info = [[Watch this space]],
}

newTalent {
  name = [[Disrupt the Continuum]],
  short_name = 'ANTI_TELE_JAM',
  display_entity = Entity.new { image='talents/disrupt_the_continuum.png' },
  type = 'skill/combat',
  mode = 'sustained',
  no_energy = true,
  info = function(who, t)
    return [[Disrupt the space-time continuum, preventing both yourself and your opponents from teleporting.]]
  end,
  sustain_describe = function(who, t, p)
    return tstring{[[You are disrupting the space-time continuum, preventing both yourself and your opponents from teleporting.  This also confers a small measure of resistance to temporal damage.]]}
  end,
  activate = function(who, def) return {} end,
  deactivate = function(who, def) return true end,
  activate = function(who, def)
    game.logPlayer(who, 'You start disrupting the magic continuum.')
    return {}
  end,
  deactivate = function(who, def)
    game.logPlayer(who, 'You stop disrupting the magic continuum.')
    return true
  end,
  apply = function(who, t)
    who.flags.NO_TELE = true
    who.flags.ANCHOR = true
    who.resist_dam[DamageType.TIME] = true
  end,
}

newTalent {
  name = [[Detect Traps]],
  short_name = 'ANTI_DET_TRAP',
  display_entity = Entity.new { image='talents/detect_traps.png' },
  type = 'skill/combat',
  mode = 'activated',
  no_energy = true,
  points = 0,
  direct_hit = true,
  action = function(self, t)
    spell.detect_traps(25, self)
    if self:getSkillLevel('ANTIMAGIC') >= 35 then
      local dam = { 0, silent_0dam=true }
      spell.ball_spell(self, t, DamageType.KILL_DOOR, dam, 1, self.x, self.y)
    end
    return true
  end,
  info = function(who, t)
    return [[Detects traps in a radius of 25 around you.  If your Antimagic skill level is 35 or higher, also destroys adjacent doors and traps.]]
  end,
}
