-- $Id: Automatizer.lua 4090 2020-09-23 00:41:19Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2019-2020 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require "engine.class"
local Dialog = require 'engine.ui.Dialog'
local ObjectData = require 'mod.class.interface.ObjectData'

module(..., package.seeall, class.make)

_M.WORLD_AUTOGEN_FILE = '/save/world-auto-auto.lua'
_M.WORLD_MANUAL_FILE = '/save/world-automatizer.lua'

_M.active = true
_M.atm_rules = {}

_M.actions = {
  destroy = true,
  pickup = true,
  inscribe = true,
  configure = true,
}

_M.test_values = {
  name = function(o, _)
    return o:getName{no_count=true, force_qty=1, no_mods=true, no_pval=true, no_inscr=true}
  end,
  inscription = function(o, _) return o.inscription or '' end,
  symbol = function(o, _) return o.display or '' end,
  type = function(o, _) return o.type or '' end,
  subtype = function(o, _) return o.subtype or '' end,
  full_type = function(o, _)
    return (o.type or '???') .. '/' .. (o.subtype or '???')
  end,
  define_as = function(o, _) return o.define_as or '' end,
  discount = function(o, _) return o:getDiscount() end,
  status = function(o, _)
    -- cf. object_status() in lib/core/auto.lua
    if not o:isIDd() then
      local pseudo_id_to_status = {
	cursed = 'bad',
	worthless = 'vary bad',
	average = 'average',
	good = 'good',
	excellent = 'very good',
	special = 'special',
	terrible = 'terrible',
      }
      return o.pseudo_id and pseudo_id_to_status[o.pseudo_id] or ''
    end
    if o:isArtifact() then return o.cursed and 'terrible' or 'special' end
    if o:isEgo() then return o.cursed and 'worthless' or 'excellent' end
    if o.type == 'weapon' then
      local to_hd = (o.to_h or 0) + (o.to_d or 0)
      return to_hd < 0 and 'bad' or to_hd > 0 and 'good' or 'average'
    end
    if o.type == 'armor' then
      local to_a = (o.to_a or 0)
      return to_a < 0 and 'bad' or to_a > 0 and 'good' or 'average'
    end
    if o.type == 'jewelry' then
      if o.pval and o.pval < 0 then return 'bad' end
      if o.subtype == 'ring' then
	local to_hd = (o.to_h or 0) + (o.to_d or 0)
	local to_a = (o.to_a or 0)
	if to_a < 0 or to_hd < 0 then return 'bad' end
      end
      return 'average'
    end
    if o.type == 'misc' and o.subtype == 'chest' then
      return o.chest.trap and 'average' or o.chest.n_obj > 0 and 'disarmed' or 'empty'
    end
    return 'average'
  end,
  race = function(_, _) return game.player.descriptor.race end,
  subrace = function(_, _) return game.player.descriptor.subrace end,
  class = function(_, _) return game.player.descriptor.class end,
  god = function(_, _) return game.player.descriptor.god end,
  descriptor = function(_, r) return game.player.descriptor[r[1]] or '' end,
  level = function(_, _) return game.player.level or 0 end,
  skill = function(_, r) return game.player:getSkillLevel(r[1]) end,
}
_M.unary_tests = {
  is_identified = function(o, _) return o:isIDd() and true or false end,
  has_ability = function(_, r)
    -- Allow ability specification by display name or short_name.
    local key = r[1]
    if game.player.abilities[key] then return true end
    for id, def in pairs(game.player.abilities_def) do
      if def.name == key and game.player.abilities[id] then return true end
    end
    return false
  end,
  has_talent = function(_, r)
    -- Allow talent specification by display name or short_name.
    local key = r[1]
    if game.player:knowTalent(key) then return true end
    for id, t in pairs(game.player.talents_def) do
      if t.name == key and game.players:knowTalent(id) then return true end
    end
    return false
  end,
  has_addon = function(_, r)
    return game:isAddonActive(r[1]) and true or false
  end,
}
_M.superclauses = {
  inventory = function(o, r, cb)
    local inv = game.player:getInven('INVEN')
    local function wrap_cb(io, _, _) return cb(r[1], io, o) end
    return util.findInInventoryBy(inv, wrap_cb) and true or false
  end,
  equipment = function(o, r, cb)
    local function wrap_cb(eo, _, inv) return inv.worn and cb(r[1], eo, o) end
    return util.findInAllInventoriesBy(game.player, wrap_cb) and true or false
  end,
}
_M.compare_ops = {
  contains = true,
  matches = true,
  oneof = true,
}

local function descriptor_names(d_type)
  local Birther = require('engine.Birther')
  local d_list = Birther.birth_descriptor_def[d_type]
  return table.maplist(function(_, d) return d.name end, d_list)
end

-- Allowed and/or meaningful values for various test clauses.
-- Primarily for use by AutomatizerMakeClause.
_M.test_allowed_values = {
  status = {
    'terrible', 'very bad', 'bad', 'average', 'good', 'very good', 'special'
  },
  race = function() return descriptor_names('race') end,
  subrace = function() return descriptor_names('subrace') end,
  class = function() return descriptor_names('class') end,
  god = function() return descriptor_names('god') end,
}

_M.tokens = {
  -- Comparison operator.
  { '^%s*([<>!=]=)', '_COMPARE_OP' },
  { '^%s*([<>])', '_COMPARE_OP' },
  -- A single punctuation character.
  { '^%s*([%(%)%{%}%-;,@])', '_LITERAL' },
  -- A single_ or double-quoted string.
  { '^%s*"(.-)"', '_STRING' },
  { "^%s*'(.-)'", '_STRING' },
  -- A number.
  { '^%s*(%d+_', '_NUMBER' },
  -- A word token.
  { '^%s*([%a_][%w_]*)', '_TOKEN' },
}

-- Scratch area for the parser.
local parser = {}

function _M:loadDefinition(file, auto, global)
  if not fs.exists(file) then return end	-- For convenience.
  local f, err = loadfile(file)
  if not f and err then error(err) end
  setfenv(f, setmetatable({
    newAutomatizerRule = function(t)
      if auto then t.auto = true end
      if global then t.global = true end
      self:newAutomatizerRule(t)
    end,
  }, {__index=_G}))
  f()
end

function _M:playerFilename(auto)
  local short_name = game.save_name:gsub('[^a-zA-Z0-9_-.]', '_'):lower()
  local save_dir = '/save/'..short_name..'/'
  local auto_sfx = auto and '-auto' or ''
  local filename = save_dir .. 'automatizer'..auto_sfx..'.lua'
  return filename, save_dir
end

function _M:loadForPlayer()
  local filename = self:playerFilename(true)
  self:loadDefinition(filename, true)
  local filename = self:playerFilename(false)
  self:loadDefinition(filename)
end

function _M:validateAutomatizerRule(t)
  assert(t.name, 'no automatizer rule name')
  assert(t.action, ('no action for automatizer rule "%s"'):format(t.name))
  assert(self.actions[t.action],
	 ('invalid action "%s" for automatizer rule "%s"'):format(t.action, t.name))
  assert(t.action ~= 'inscribe' or type(t.inscription) == 'string',
	 ('missing or invalid inscription "%s" for automatizer rule "%s"'):format(tostring(t.inscription), t.name))
  assert(t.action == 'inscribe' or not t.inscription,
	 ('spurious inscription for %s automatizer rule "%s"'):format(t.action, t.name))
  if t.action == 'configure' then
    assert(type(t.config) == 'table',
	   ('missing or malformeed config{} table for automatizer rule "%s"'):format(t.name))
    for k, v in pairs(t.config) do
      assert(ObjectData.config_types[k] and type(k) == 'string' and type(v) == 'boolean',
	     ('missing or malformeed config{} action "%s"=%s for automatizer rule "%s"'):format(tostring(k), tostring(v), t.name))
    end
  else
    assert(not t.config, ('spurious config{} table for %s automatizer rule "%s"'):format(t.action, t.name))
  end
  assert(t.match, ('no match for automatizer rule "%s"'):format(t.name))
  assert(not self.atm_rules[t.name], ('automatizer rule collision on name "%s"'):format(t.name))
  if type(t.match) == 'string' then
    t.match = self:parseRuleText(t)
    if not t.match then return end
  else
    -- TODO
    -- local ok, err = self:isValidMatch(t.match, false)
  end
end

function _M:newAutomatizerRule(t)
  self:validateAutomatizerRule(t)
  table.insert(self.atm_rules, t)
  self.atm_rules[t.name] = t
end

function _M:getRule(name)
  return self.atm_rules[name]
end

function _M:decollideRuleName(name, rules)
  rules = rules or self.atm_rules
  local new_name, i = name, 0
  while rules[new_name] do
    i = i + 1
    new_name = name .. ' #' .. i
  end
  return new_name
end

function _M:itemMatches(rule, o, oo)
  -- If 'rule' is the actual rule definition, use its match{} subtable.
  if rule.match then return self:itemMatches(rule.match, o, oo) end
  if rule.test == 'placeholder' then return false end
  if rule.test == 'and' then
    for _, cl in ipairs(rule) do
      if not self:itemMatches(cl, o, oo) then return false end
    end
    return true
  end
  if rule.test == 'or' then
    for _, cl in ipairs(rule) do
      if self:itemMatches(cl, o, oo) then return true end
    end
    return false
  end
  if rule.test == 'not' then return not self:itemMatches(rule[1], o, oo) end
  if rule.test == '==' then
    return self:exprVal(rule[1], o, oo) == self:exprVal(rule[2], o, oo)
  end
  if rule.test == '!=' then
    return self:exprVal(rule[1], o, oo) ~= self:exprVal(rule[2], o, oo)
  end
  if rule.test == '<' then
    return self:exprVal(rule[1], o, oo) < self:exprVal(rule[2], o, oo)
  end
  if rule.test == '>' then
    return self:exprVal(rule[1], o, oo) > self:exprVal(rule[2], o, oo)
  end
  if rule.test == '<=' then
    return self:exprVal(rule[1], o, oo) <= self:exprVal(rule[2], o, oo)
  end
  if rule.test == '>=' then
    return self:exprVal(rule[1], o, oo) >= self:exprVal(rule[2], o, oo)
  end
  if rule.test == 'inrange' then
    local val = self:exprVal(rule[1], o, oo)
    return val >= self:exprVal(rule[2], o, oo) and val <= self:exprVal(rule[3], o, oo)
  end
  if rule.test == 'oneof' then
    local val = self:exprVal(rule[1], o, oo)
    for i = 2, #rule do
      if val == self:exprVal(rule[i], o, oo) then return true end
    end
    return false
  end
  if rule.test == 'contains' then
    return self:exprVal(rule[1], o, oo):lower():find(self:exprVal(rule[2], o, oo), 1, true) and true or false
  end
  if rule.test == 'matches' then
    return self:exprVal(rule[1], o, oo):lower():find(self:exprVal(rule[2], o, oo)) and true or false
  end
  -- If we're not a test clause, we may be a unitary boolean value.
  local val = self:exprVal(rule, o, oo)
  if type(val) == 'boolean' then return val end
  -- We should probably whinge about a malformed rule here.
  return false
end

function _M:exprVal(clause, o, oo)
  if type(clause) == 'number' or type(clause) == 'string' then return clause end
  -- This will probably choke upstream.
  if type(clause) ~= 'table' or not clause.clause then return nil end
  if self.superclauses[clause.clause] then
    local function cb(rule, xo, xoo)
      return self:itemMatches(rule, xo, xoo)
    end
    return self.superclauses[clause.clause](o, clause, cb)
  end
  local f = self.test_values[clause.clause] or self.unary_tests[clause.clause]
  if f then
    if clause.base and not oo then return nil end
    local o_use = clause.base and oo or o
    return f(o_use, clause)
  end
  return nil
end

function _M:checkInventory(who)
  if not self.active then return end
  local inv = who:getInven('INVEN')
  for i = #inv, 1, -1 do
    for _, rule in ipairs(self.atm_rules) do
      if self:itemMatches(rule, inv[i]) then
	if rule.action == 'destroy' then
	  if not inv[i]:isArtifact() and not inv[i].quest_obj then
	    local o = who:removeObject('INVEN', i, true)
	    game.log('#LIGHT_RED#Automatizer destroys:#LAST#  '..o:getName())
	    if game:hasEntity(o) then game:removeEntity(o) end
	    who.changed = true
	    break
	  else
	    game.log('#LIGHT_RED#Automatizer cannot destroy:#LAST#  '..o:getName())
	    if not o:isIDd() then
	      o:forAllStack(function(so)
		o.pseudo_id = o.cursed and 'terrible' or 'special'
	      end)
	    end
	  end
	elseif rule.action == 'inscribe' then
	  inv[i]:forAllStack(function(o)
	    o.inscription = rule.inscription
	  end)
	elseif rule.action == 'configure' then
	  inv[i]:forAllStack(function(o)
	    for k, v in pairs(rule.config) do
	      if v then o:setConfig(k, true) end
	    end
	  end)
	end
      end
    end
  end
end

function _M:checkFloor(x, y)
  if not self.active then return end
  local map = game.level.map
  for i = map:getObjectTotal(x, y), 1, -1 do
    local o = map:getObject(x, y, i)
    for _, rule in ipairs(self.atm_rules) do
      if o and self:itemMatches(rule, o) then
	if rule.action == 'destroy' then
	  if not o:isArtifact() and not o.quest_obj then
	    map:removeObject(x, y, i)
	    game.log('#LIGHT_RED#Automatizer destroys:#LAST#  '..o:getName())
	    if game:hasEntity(o) then game:removeEntity(o) end
	    break
	  else
	    game.log('#LIGHT_RED#Automatizer cannot destroy:#LAST#  '..o:getName())
	    if not o:isIDd() then
	      o:forAllStack(function(so)
		o.pseudo_id = o.cursed and 'terrible' or 'special'
	      end)
	    end
	  end
	elseif rule.action == 'pickup' then
	  who:pickupFloor(i, true)
	  who.changed = true
	  break
	end
      end
    end
  end
end

-- Extract a lexical token from the working input string.  If a token has
-- already been read previously and not yet consumed by a call to match()
-- or maybe_match(), that token is returned instead.  Sets parser.tok to
-- the text of the token, and parser.toktype to the type of the token.
-- Returns the cached token, or the next token read from input, or nil at
-- end of file.
local function token()
  -- If we've already matched a token, return it.
  if parser.tok then return parser.tok end

  -- Try all the token types we recognize.
  for _, token in ipairs(_M.tokens) do
    local i, j, tok = parser.text:find(token[1])
    if i and j then
      parser.tok = tok
      parser.toktype = token[2]
      parser.text = parser.text:sub(j+1)
      return parser.tok
    end
  end

  -- No match; assume end of string.
end

-- Returns true if the next input token (as returned by token()) matches
-- any of the tokens or token types passed as arguments.  look('EOF')
-- returns true at end of file
local function look(...)
  local args = {...}
  parser.tok = parser.tok or token()
  -- look() always returns true, so that match() always returns the current
  -- token.
  if #args == 0 then return true end
  if #args == 1 and args[1] == 'EOF' then return not parser.tok end
  for _, arg in ipairs(args) do
    if arg == parser.tok or arg == parser.toktype then return true end
  end
  return false
end

-- Returns true if the next input token (as returned by token()) matches
-- (as per look()) any of the keys of the provided hash table argument.
-- Typically this will be used with one of the internal matching tables
-- like _M.test_values{}.
local function look_token(t)
  parser.tok = parser.tok or token()
  for token, _ in pairs(t) do
    if token == parser.tok then return true end
  end
  return false
end

-- Consumes and returns the next token (as returned by token()) if it
-- matches any of the provided tokens or token types (as per look());
-- otherwise, leaves the token unconsumed and returns nil.
local function maybe_match(...)
  if not look(...) then return nil end
  local ret = parser.tok
  parser.tok = nil
  parser.toktype = nil
  return ret
end

-- Consumes and returns the next token (as returned by token()) if it
-- matches (as per look()) any of the keys of the provided hash table
-- argument.  Typically this will be used with one of the internal matching
-- tables like _M.test_values{}.  If no match is found, leaves the token
-- unconsumed and returns nil.
local function maybe_match_token(t)
  if not look_token(t) then return nil end
  local ret = parser.tok
  parser.tok = nil
  parser.toktype = nil
  return ret
end

-- Consumes and returns the next token (as returned by token()) if it
-- matches any of the provided tokens or token types (as per look());
-- otherwise, throws an error.
local function match(...)
  if look(...) then
    local ret = parser.tok
    parser.tok = nil
    parser.toktype = nil
    return ret
  end

  local args = {...}
  args = table.concat(args, ", ")
  local found = parser.tok and "'"..parser.tok..'"' or 'end of file'
  error(('expected one of { %s } instead of %s'):format(args, found))
  return nil
end

-- Matches and consumes an optional list of arguments for a clause, in the
-- form of a comma-separated list of number or string values in
-- parentheses.  Returns a list of the found values; returns an empty list
-- if no clause arguments are present.
local function clause_args()
  local ret = {}
  if maybe_match('(') then
    ret[#ret+1] = match('_NUMBER', '_STRING')
    while maybe_match(',') do ret[#ret+1] = match('_NUMBER', '_STRING') end
    match(')')
  end
  return ret
end

-- Matches and consumes an optional "base object" marker inside a super-
-- clause.  Returns true if a base marker was present.
local function base_marker()
  if not maybe_match('@') then return nil end
  if not parser.in_superclause then
    error("Cannot use '@' marker outside 'inventory{}' or 'equipment{}'")
    return nil
  end
  return true
end

-- Matches and consumes a simple "value" expression:  either a number or
-- string or one of the testable values defined in _M.test_values{} with an
-- optional argument list [cf. clause_args()].  Returns an automatizer data
-- structure representing the expression.
local function value()
  local ret = maybe_match('_NUMBER', '_STRING')
  if ret then return ret end

  if look_token(_M.test_values) then
    local cl = match()
    local is_base = base_marker()
    local ret = clause_args()
    ret.clause = cl
    ret.base = is_base
    return ret
  end
  error(('Expected expression before "%s"'):format(tostring(parser.tok)))
end

-- Matches and consumes a single conditional clause, possibly preceded by
-- 'not', or a parenthesized list of such clauses.  Returns an automatizer
-- data structure representing the clause.
local function clause()
  -- Catch 'not' and parenthesized expressions first
  if maybe_match('(') then
    local ret = _M:clauses('or')
    match(')')
    return ret
  end
  if maybe_match('not') then
    return { test='not', clause() }
  end

  -- Catch "super-clauses".
  if look_token(_M.superclauses) then
    local cl = match()
    if parser.in_superclause then
      -- No nested superclauses; that would get too hairy.
      error(("Cannot use '%s{}' inside '%s{}'"):format(cl, parser.in_superclause))
      return nil
    end
    match('{')
    parser.in_superclause = cl
    local subcl = clause()
    parser.in_superclause = nil
    match('}')
    return { clause=cl, subcl }
  end

  -- Check for a unary condition.
  if look_token(_M.unary_tests) then
    local cl = match()
    local is_base = base_marker()
    local ret = clause_args()
    ret.clause = cl
    ret.base = is_base
    return ret
  end

  -- Everything else should be a binary comparison between two values (with
  -- exceptions like 'oneof' or 'inrange'
  local lhs = value()
  local op = maybe_match('_COMPARE_OP') or maybe_match_token(_M.compare_ops)
  if not op then
    error('expected comparison operator')
    return nil
  end
  if op == 'inrange' then
    local lb = match('_NUMBER')
    match(',')
    local rb = match('_NUMBER')
    return { test=op, lhs, lb, rb }
  end
  if op == 'oneof' then
    local ret = { test=op, lhs, match('_NUMBER', '_STRING') }
    while maybe_match(',') do
      ret[#ret+1] = match('_NUMBER', '_STRING')
    end
    return ret
  end
  if op == 'contains' or op == 'matches' then
    return { test=op, lhs, match('_STRING') }
  end
  local rhs = value()
  return { test=op, lhs, rhs }
end

-- Matches and consumes one or more clauses separated by 'and', or one or
-- more groups of 'and'-ed clauses separated by 'or', as per the provided
-- join parameter.  Returns an automatizer data structure representing the
-- combined clauses.
function _M:clauses(join)
  local ret = { test=join }
  ret[#ret+1] = join == 'or' and self:clauses('and') or clause()
  while maybe_match(join) do
    ret[#ret+1] = join == 'or' and self:clauses('and') or clause()
  end
  return #ret == 1 and ret[1] or ret
end

-- TODO
function _M:isValidMatch(match, in_superclause)
  if match.test then
    if match.test == 'and' or match.test == 'or' then
      for _, subcl in ipairs(match) do
	local ok, err = self:isValidMatch(subcl, in_superclause)
	if not ok then return ok, err end
      end
      return true
    end
    if match.test == 'oneof' then
      if #match < 2 then
	return false, "no candidates for 'match' clause"
      end
      for i = 2, #match do
	if type(match[i]) ~= 'string' and type(match[i]) ~= 'number' then
	  return false, "non-constant candidate in 'oneof' clause"
	end
      end
      return self:isValidMatch(match[1], in_superclause)
    end
    if match.test == 'inrange' then
      if #match ~= 3 then
	return false, ("wrong number of bounds for 'inrange' clause (need 2, found %d)"):format(#match - 1)
      end
      if type(match[2]) ~= 'number' or type(match[3]) ~= 'number' then
	return false, "non-numeric bound in 'inrange' clause"
      end
      if match[2] > match[3] then
	return false, ("crossed bounds in 'inrange' clause (%d > %d)"):format(match[2], match[3])
      end
      return self:isValidMatch(match[1], in_superclause)
    end
    if #match ~= 2 then
      return false, ("wrong number of operands for '%s' clause (need 2, found %d)"):format(match.test, #match)
    end
    if match.test == 'contains' or match.test == 'matches' then
      if type(match[2]) ~= 'string' then
	return false, ("non-string constraint for '%s' clause"):format(match.test)
      end
      return self:isValidMatch(match[1], in_superclause)
    end
  end
  if match.clause then
    -- TODO
  end
  return true
end

function _M:parseRuleText(t)
  parser.tok = nil
  parser.toktype = nil
  parser.text = t.match
  local ok, ret = pcall(function() return self:clauses('or') end)
  if ok then return ret end
  Dialog:simpleLongPopup('Automatizer Error', ('Error parsing automatizer rule "%s":\n%s'):format(t.name or '???', tostring(ret)), 600)
  return nil
end

function _M:writeRule(rule, f)
  local function write_match(m, ind)
    if type(m) == 'string' then return ('%q'):format(m) end
    if type(m) == 'number' then return ('%d'):format(m) end
    if type(m) ~= 'table' then return tostring(m) end

    local ret
    if m.test then
      ret = ("{ test='%s'"):format(m.test)
    elseif m.clause then
      ret = ("{ clause='%s'"):format(m.clause)
      if m.base then ret = ret .. ', base=true' end
    else
      ret = '{ --[[???]] '
    end
    for _, mm in ipairs(m) do
      ret = ret .. (type(mm) == 'table' and (',\n  '..ind) or ', ')
      ret = ret .. write_match(mm, ind .. '  ')
    end
    ret = ret .. (type(m[#m]) == 'table' and ('\n'..ind..'}') or ' }')
    return ret
  end

  f:write('\nnewAutomatizerRule {\n')
  f:write(('  name = %q,\n'):format(rule.name))
  f:write(("  action = '%s',\n"):format(rule.action))
  if rule.inscription and rule.action == 'inscribe' then
    f:write(('  inscription = %q,\n'):format(rule.inscription))
  end
  if rule.config and rule.action == 'configure' then
    f:write('  config = {\n')
    for tag, v in pairs(rule.config) do
      if v then f:write(('    %s = true,\n'):format(tag)) end
    end
    f:write('  },\n')
  end
  f:write(('  match = %s,\n'):format(write_match(rule.match, '  ')))
  -- Note that we don't write out the 'auto' or 'global' fields; those are
  -- determined by which definitions file they're loaded from (and in turn
  -- determine which definitions file they'll be written to).
  f:write('}\n')
end

function _M:writeAutoRules()
  local hdr = [[-- Auto-generated automatizer rules and rules created via the in-game
-- Automatizer Rules dialog.  You generally should not edit this file
-- directly unless you know *exactly* what you're doing; if you want to
-- write your own hand-crafted automatizer rules, you can place them in the
-- file 'world-automatizer.lua' in the save/ directory (next to
-- world.teaw), or the file 'automatizer.lua' in a particular character's
-- save directory.  See <insert URL here> for information about the ATM
-- rule description language.
--
]]

  -- Perhaps overkill, but just in case...
  fs.mkdir('/save/')
  local f = fs.open(self.WORLD_AUTOGEN_FILE, 'w')
  f:write(hdr)
  f:write('-- Auto-generated rules that apply to all characters.\n')
  for _, rule in ipairs(self.atm_rules) do
    if rule.auto and rule.global then self:writeRule(rule, f) end
  end
  f:close()

  local filename, save_dir = self:playerFilename(true)
  fs.mkdir(save_dir)
  f = fs.open(filename, 'w')
  f:write(hdr)
  f:write('-- Auto-generated rules that apply to '..game.save_name..'.\n')
  for _, rule in ipairs(self.atm_rules) do
    if rule.auto and not rule.global then self:writeRule(rule, f) end
  end
  f:close()
end
