-- $Id: Game.lua 4095 2020-09-26 21:48:08Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2020 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require "engine.class"
require "engine.GameTurnBased"
require "engine.interface.GameTargeting"
require "engine.KeyBind"
local Savefile = require "engine.Savefile"
local DamageType = require "engine.DamageType"
local Zone = require "mod.class.Zone"
local Tiles = require "engine.Tiles"
local Map = require "mod.class.Map"
local Level = require "engine.Level"
local Birther = require "mod.dialogs.Birther"
local GameState = require 'mod.class.GameState'
local GameOptions = require 'mod.dialogs.GameOptions'
local Shader = require 'engine.Shader'

local Object = require 'mod.class.Object'
local Grid = require "mod.class.Grid"
local Store = require 'mod.class.Store'
local StoreOwner = require 'mod.class.StoreOwner'
local Home = require 'mod.class.Home'
local MathomHouse = require 'mod.class.MathomHouse'
local Actor = require "mod.class.Actor"
local Player = require "mod.class.Player"
local NPC = require "mod.class.NPC"

local Dialog = require 'engine.ui.Dialog'
local HotkeysDisplay = require "mod.class.HotkeysDisplay"
local HotkeysIconsDisplay = require "mod.class.HotkeysIconsDisplay"
local ActorsSeenDisplay = require "engine.ActorsSeenDisplay"
local LogDisplay = require "mod.class.LogDisplay"
local LogFlasher = require "mod.class.LogFlasher"
local StatsDisplay = require 'mod.class.StatsDisplay'
local DebugConsole = require "engine.DebugConsole"
local FlyingText = require "engine.FlyingText"
local Tooltip = require "engine.Tooltip"
local BigNews = require 'mod.class.BigNews'
local FontPackage = require 'engine.FontPackage'
local VisualsDialog = require 'mod.dialogs.VisualsDialog'

local Calendar = require "engine.Calendar"
local Automatizer = require 'mod.class.Automatizer'

local UseItemDialog = require 'mod.dialogs.UseItemDialog'

local QuitDialog = require "mod.dialogs.Quit"

module(..., package.seeall, class.inherit(engine.GameTurnBased, engine.interface.GameTargeting))

function _M:init()
  engine.GameTurnBased.init(self, engine.KeyBind.new(), 1000, 100)

  -- Pause at birth
  self.paused = true

  -- Same init as when loaded from a savefile
  self:loaded()
end

function _M:run()
  class:triggerHook{ 'T2:runBegin' }

  self.calendar = Calendar.new('/data/calendar.lua', 'Today is the %s %s of the %s year of the third age.\nThe time is %02d:%02d', 2890, 130, 6)

  Automatizer:loadDefinition(Automatizer.WORLD_AUTOGEN_FILE, true, true)
  Automatizer:loadDefinition(Automatizer.WORLD_MANUAL_FILE, false, true)

  local font_mono, size_mono = FontPackage:getFont("mono_small", "mono")
  local cfg = config.settings.tome2

  self.flash = LogFlasher.new(0, 0, self.w, 20, nil, nil, nil, {255,255,255}, {0,0,0})
  self.logdisplay = LogDisplay.new(0, self.h * 0.8, self.w * 0.5, self.h * 0.2, nil, nil, nil, {255,255,255}, {30,30,30})
  --self.hotkeys_display_text = HotkeysDisplay.new(nil, self.w * 0.5, self.h * 0.8, self.w * 0.5, self.h * 0.2, {30,30,0})
  self.npcs_display = ActorsSeenDisplay.new(nil, self.w * 0.5, self.h * 0.8, self.w * 0.5, self.h * 0.2, {30,30,0})
  self.stats_display = StatsDisplay.new(nil, 0, 20, 200, self.h * 0.8 - 20, nil, nil, 14)
  self.tooltip = Tooltip.new(nil, nil, {255,255,255}, {30,30,30})
  self.flyers = FlyingText.new()
  self:setFlyingText(self.flyers)

  self.bignews = BigNews.new(FontPackage:getFont('bignews'))

  self.particles = {}

  self.log = function(style, ...)
    if type(style) == "number" then
      self.logdisplay(...)
      self.flash(style, ...)
    else
      self.logdisplay(style, ...)
      self.flash(self.flash.NEUTRAL, style, ...)
    end
  end
  self.logSeen = function(e, style, ...)
    if e and self.level.map.seens(e.x, e.y) then
      self.log(style, ...)
    end
  end
  self.logSeenXY = function(x, y, style, ...)
    if self.level.map.seens(x, y) then
      self.log(style, ...)
    end
  end
  self.logPlayer = function(e, style, ...)
    if e == self.player then
      self.log(style, ...)
    end
  end
  self.logChat = function(msg, ...)
    self.log(msg, ...)
  end
  self.logRollback = function(line, ...)
    return self.logdisplay:rollback(line, ...)
  end
  self.logNewest = function()
    return self.logdisplay:getNewestLine()
  end

  self.log(self.flash.GOOD, "Welcome to #00FF00#Tales of Middle-Earth#LAST#!")

  --self.hotkeys_display_text:setColumns(5)
  self:resizeIconsHotkeysToolbar()

  -- Setup inputs
  self:setupCommands()
  self:setupMouse()

  self:setupDisplayMode('postinit')

  -- Starting from here we create a new game
  if not self.player then
    self:newGame()
  else
    Automatizer:loadForPlayer()
  end

  self.hotkeys_display.actor = self.player
  self.npcs_display.actor = self.player
  self.stats_display.actor = self.player

  -- Setup the targetting system
  engine.interface.GameTargeting.init(self)

  -- Create scroll mode text overlay
  local lfont = FontPackage:get('bignews', true)
  lfont:setStyle('bold')
  local s = core.display.drawStringBlendedNewSurface(lfont, '[Use direction keys to scroll, <CapsLock> to exit]', unpack(colors.simple(colors.GOLD)))
  lfont:setStyle('normal')
  self.caps_scroll = { s:glTexture() }
  self.caps_scroll.w, self.caps_scroll.h = s:getSize()

  -- Ok everything is good to go, activate the game in the engine!
  self:setCurrent()

  class:triggerHook{ 'T2:runReady' }
end

function _M:setPlayerName(name)
  self.save_name = name
  self.player_name = name
  if self.player then self.player.name = name end
end

function _M:newGame()
  self.player = Player.new{name=self.player_name, game_ender=true}
  Map:setViewerActor(self.player)

  self.state = GameState.new{}
  self.always_target = true

  self.creating_player = true
  local birth = Birther.new(nil, self.player, {'base', 'sex', 'race', 'subrace', 'class_group', 'class', 'god'}, function()
    self.calendar = Calendar.new('/data/calendar.lua', 'Today is the %s %s of the %s year of the third age.\nThe time is %02d:%02d.', 2890, 130, 6)
    -- Vaguely Convoluted Hack(TM):  Before we can create the Bree level,
    -- we have to know where Bree is on the map.
    local wxy = self.state:getWorldMapZoneCoords('town-bree')
    if wxy then self.player.wild_x, self.player.wild_y = wxy.x, wxy.y end
    self:changeLevel(1, "town-bree")
    self.player:resolve()
    self.player:resolve(nil, true)
    self.player:recalcEverything()
    self.player:recalcAppearance()
    self.player.energy.value = self.energy_to_act
    self.paused = true
    self.creating_player = false
    Automatizer:loadForPlayer()
    self.player:grantQuest(self.player.starting_quest)
    Map:setViewerFaction(self.player.faction)
    if config.settings.cheat then self.player.__cheated = true end
  end)
  self:registerDialog(birth)
end

function _M:loaded()
  engine.GameTurnBased.loaded(self)
  Zone:setup {
    map_class = 'mod.class.Map',
    npc_class = "mod.class.NPC",
    grid_class = "mod.class.Grid",
    object_class = 'mod.class.Object',
    trap_class = 'mod.class.Trap',
  }
  Map:setViewerActor(self.player)
  self:setupDisplayMode('init')
  self:setupDisplayMode('postinit')
  self.key = engine.KeyBind.new()

  if self.player then
    self.player:recalcAppearance()
    Map:setViewerFaction(self.player.faction)
  end
  if self.player and config.settings.cheat then
    self.player.__cheated = true
  end
end

function _M:setupDisplayMode(mode)
  local cfg = config.settings.tome2

  if not mode or mode == 'init' then
    Map:resetTiles()
  end

  if not mode or mode == 'postinit' then
    local tileset = cfg.tiles
    Tiles.prefix = '/data/gfx/'..tileset..'/'

    local size_opt = tileset == 'ascii' and 'ascii_tile_size' or 'tile_size'
    local def = GameOptions.options_def[size_opt]
    local ts = cfg[size_opt] or def.default
    local tsw, tsh
    if def.px_sizes then
      tsw, tsh = unpack(def.px_sizes[ts])
    else
      tsw, tsh = def.px_size[ts], def.px_size[ts]
    end

    local map_w = math.floor((self.w - 200)/tsw)*tsw
    local map_h = math.floor((self.h*0.80 - 20)/tsh)*tsh
    -- Annoyingly, the map viewport is coming out larger than we're
    -- requesting; pull it in some to compensate.
    Map:setViewPort(self.w - map_w, 20, map_w - 2*tsw, map_h - 2*tsh, tsw, tsh, nil, tsh-2, true)

    if tileset == 'ascii' then
      Map.tiles.use_images = false
      Map.tiles.force_back_color = { r=0, g=0, b=0, a=255 }
    else
      Map.tiles.use_images = true
      Map.tiles.sharp_scaling = true
    end
    Map.object_stack_count = true

    if self.level then
      if self.level.map.finished then
	VisualsDialog:updateLevel()
	self.level.map:recreate()
	self.level.map:moveViewSurround(self.player.x, self.player.y, 8, 8)
      end
      engine.interface.GameTargeting.init(self)
    end

    self:createFBOs()
  end
end

function _M:createFBOs()
  print('[GAME] Creating FBOs')

  self.fbo = core.display.newFBO(Map.viewport.width, Map.viewport.height)
  if self.fbo then
    self.fbo_shader = Shader.new('main_fbo')
    self.posteffects = {
      wobbling	= Shader.new('main_fbo/wobbling'),
      blur	= Shader.new('main_fbo/blur'),
    }
    self.posteffects_use = { self.fbo_shader.shad }
    if not self.fbo_shader.shad then
      self.fbo = nil
      self.fbo_shader = nil
    end
    self.fbo2 = core.display.newFBO(Map.viewport.width, Map.viewport.height)
  end

  if self.player then self.player:updateMainShader() end

  self.full_fbo = core.display.newFBO(self.w, self.h)
  if self.full_fbo then
    self.full_fbo_shader = Shader.new('full_fbo')
    if not self.full_fbo_shader.shad then
      self.full_fbo = nil
      self.full_fbo_shader = nil
    end
  end

  if self.fbo and self.fbo2 then
    core.particles.defineFramebuffer(self.fbo)
  else
    core.particles.defineFramebuffer(nil)
  end

  if self.target then
    self.target:enableFBORenderer('targetshader.png', 'target_fbo')
  end
  Map:enableFBORenderer('target_fbo')
end

function _M:resizeIconsHotkeysToolbar()
  local cfg = config.settings.tome2
  local size = cfg.hotkey_icons_size
  local fontsize = math.floor(size/4)

  if self.hotkeys_display_icons then
    if fontsize ~= self.hotkeys_display_icons.fontsize then
      -- Adjust the font size with the icon size.
      local toolbar = self.hotkeys_display_icons
      local fontname = toolbar.fontname or "/data/font/DroidSansMono.ttf"
      toolbar.fontsize = fontsize
      toolbar.font = core.display.newFont(fontname, fontsize)
      toolbar.fontbig = core.display.newFont(fontname, fontsize*2)
      toolbar.font_h = toolbar.font:lineSkip()
    end
    self.hotkeys_display_icons:resize(self.w * 0.5, self.h * 0.8, self.w * 0.5, self.h * 0.2, size, size)
  else
    self.hotkeys_display_icons = HotkeysIconsDisplay.new(nil, self.w * 0.5, self.h * 0.8, self.w * 0.5, self.h * 0.2, {112,128,144}, font_mono, fontsize, size, size)
    self.hotkeys_display_icons:enableShadow(0.6)
  end

  self.hotkeys_display = self.hotkeys_display_icons
  self.hotkeys_display.actor = game.player
end

function _M:save()
  return class.save(self, self:defaultSavedFields{on_level_load_fcts=true}, true)
end

function _M:isTainted()
  if config.settings.cheat then return true end
  return (game.player and game.player.__cheated) and true or false
end

function _M:getSaveDescription()
  local pl = self.player
  local race = pl.descriptor.race
  local subrace = pl.descriptor.subrace
  local class = pl.descriptor.class
  if subrace ~= 'Base' then
    local srd = Birther:getBirthDescriptor('subrace', subrace)
    race = srd.name_mod == 'suffix' and race..' '..subrace or subrace..' '..race
  end
  local name = ('%s the level %d %s %s'):format(pl.name, pl.level, race, class)
  local lev = self.zone:level_adjust_level(self.level, 'actor')
  local loc = ('Currently on level %d of %s'):format(lev, self.zone.name)
  if self.zone.max_level == 1 then loc = 'Currently in '..self.zone.name end
  return {
    name = self.player.name,
    description = name.."\n"..loc,
  }
end

function _M:leaveLevel(level, lev, old_lev)
  if level:hasEntity(self.player) then
    level.exited = level.exited or {}
    if lev > old_lev then
      level.exited.down = {x=self.player.x, y=self.player.y}
    else
      level.exited.up = {x=self.player.x, y=self.player.y}
    end
    level.last_turn = game.turn
    level:removeEntity(self.player)
  end
end

function _M:onLevelLoad(id, fct, data)
  if self.zone and self.level and id == self.zone.short_name.."-"..self.level.level then
    print("Direct execute of on level load", id, fct, data)
    fct(self.zone, self.level, data)
    return
  end

  self.on_level_load_fcts = self.on_level_load_fcts or {}
  self.on_level_load_fcts[id] = self.on_level_load_fcts[id] or {}
  local l = self.on_level_load_fcts[id]
  l[#l+1] = {fct=fct, data=data}
  print("Registering on level load", id, fct, data)
end

function _M:onLevelLoadRun()
  self.on_level_load_fcts = self.on_level_load_fcts or {}
  local id = self.zone.short_name.."-"..self.level.level
  print("Running on level loads", id)
  for i, fct in ipairs(self.on_level_load_fcts[id] or {}) do
    fct.fct(self.zone, self.level, fct.data)
  end
  self.on_level_load_fcts[id] = nil
end

local level_feelings = {
  [0] =  "Looks like any other level.",
  [1] =  "You feel there is something special about this level.",
  [2] =  "You have a superb feeling about this level.",
  [3] =  "You have an excellent feeling...",
  [4] =  "You have a very good feeling...",
  [5] =  "You have a good feeling...",
  [6] =  "You feel strangely lucky...",
  [7] =  "You feel your luck is turning...",
  [8] =  "You like the look of this place...",
  [9] =  "This level can't be all bad...",
  [10] = "What a boring place..."
}

function _M:changeLevel(lev, zone, params)
  params = params or {}

  -- Extremely Annoying Hack(TM):  Catch departure from one-time special
  -- levels at the last possible moment, since we lack a blocking dialog to
  -- do it from anywhere else...
  local flag, ok = table.get(self, 'level', 'data', 'generator', 'map', 'one_time_level')
  if ok and flag and not params.force_one_time_level then
    local function ok_cb(ok)
      if ok then
	params.force_one_time_level = true
	self:changeLevel(lev, zone, params)
      end
    end
    local title = self.level.data.level_name or 'Special Level'
    Dialog:yesnoPopup(title, 'Leave this special level forever?', ok_cb)
    return
  end
  -- Extremely Annoying Hack(TM), redux and epilogue:  Clean this up behind
  -- ourself.
  params.force_one_time_level = nil

  local old_zone = self.zone
  local old_lev = (self.level and not zone) and self.level.level or -1000
  -- If we're leaving the worldmap, we'll need to check our light radius.
  local leaving_worldmap = zone and self.zone and self.zone.worldmap
  local redo_lite = leaving_worldmap

  if self.zone and self.level then
    self.player:fireCallbacks('on_leave_level', self.zone, self.level, old_lev, params)
  end

  -- If we're recalling out, remember where we're recalling from.
  if params.recall then game.state:setRecallDest() end

  if zone then
    if util.zone_is(zone, 'worldmap') then
      -- If we're entering the worldmap from another zone, try to set our
      -- wilderness x/y based on the zone we're leaving.
      local zxy = self.state:getWorldMapZoneCoords(params.use_zone or self.zone.short_name)
      if zxy then self.player.wild_x, self.player.wild_y = zxy.x, zxy.y end
      -- And we'll need to check our lite radius.
      redo_lite = true
    end

    if self.zone then
      if self.zone.on_leave then
	self.zone:on_leave(lev, old_lev, zone, params)
      end
      self.zone:leaveLevel(false, lev, old_lev)
      self.zone:leave()
    end

    if type(zone) == "string" then
      self.zone = Zone.new(zone)
    else
      self.zone = zone
    end
  end

  self.zone:getLevel(self, lev, old_lev)
  self.zone:runPostGeneration(self.level)
  self:onLevelLoadRun()

  -- Slight Hack(TM):  If we're in a persistent zone, re-apply the visuals
  -- settings, in case the player has changed them since last visit.
  if game.zone.persistent == 'zone' then
    VisualsDialog:updateLevel()
  end

  local startxy = self.zone.entry_point and self.zone:entry_point(lev, old_lev, old_zone, params) or nil
  if startxy then
    if startxy.use and self.level[startxy.use] then
      -- entry_point() returned { use='default_up'/'default_down' }
      local use = self.level[startxy.use]
      print(('[GAME] changeLevel:  moving to level %s (%d,%d) per zone instruction'):format(startxy.use, use.x, use.y))
      self.player:move(use.x, use.y, true)
    elseif startxy.type and startxy.subtype then
      -- entry_point() returned { type='something', subtype='something' }
      local spot = self.level:pickSpot(startxy)
      if spot then
	print(('[GAME] changeLevel:  moving to %s/%s level spot (%d,%d)'):format(startxy.type, startxy.subtype, spot.x, spot.y))
	self.player:move(spot.x, spot.y, true)
      else
	error(('[GAME] changeLevel:  no %s/%s level spot; falling back to default_up (%d,%d)'):format(startx.type, startx.subtype, self.level.default_up.x, self.level.default_up.y))
	self.player:move(self.level.default_up.x, self.level.default_up.y, true)
      end
    else
      -- entry_point() returned { x=#, y=# }
      print(('[GAME] changeLevel:  moving to zone entry point (%d,%d)'):format(startxy.x, startxy.y))
      self.player:move(startxy.x, startxy.y, true)
    end
  elseif lev >= old_lev and self.level.default_up.x and self.level.default_up.y then
    -- We use >= above to handle spell.alter_reality(), which will call us
    -- with lev == old_lev.
    print(('[GAME] changeLevel:  moving to level default_up (%d,%d)'):format(self.level.default_up.x, self.level.default_up.y))
    self.player:move(self.level.default_up.x, self.level.default_up.y, true)
  elseif self.level.default_down.x and self.level.default_down.y then
    print(('[GAME] changeLevel:  moving to level default_down (%d,%d)'):format(self.level.default_down.x, self.level.default_down.y))
    self.player:move(self.level.default_down.x, self.level.default_down.y, true)
  else
    -- ??? Emergency fallback:  just pick a random suitable spot.
    print('[GAME] changeLevel:  ??? trying to find a suitable entry loc')
    while true do
      local x = rng.range(1, self.level.map.w - 1)
      local y = rng.range(1, self.level.map.h - 1)
      if self.level.map:checkEntity(x, y, Map.TERRAIN, 'empty_bold') then
	print(('[GAME] changeLevel:      placing at %d,%d'):format(x, y))
	self.player:move(x, y, true)
	break
      end
      print(('[GAME] changeLevel:      %d,%d not suitable'):format(x, y))
    end
  end
  self.level:addEntity(self.player)

  if self.zone.on_enter then
    self.zone:on_enter(lev, old_lev, old_zone, params)
  end
  if self.level.data.generator.map.one_time_level then
    -- One-time-only special level; remember that we've done it.
    game.state:doneSpecialLevel(self.level.data.generator.map.one_time_level, true)
  end

  -- Enforce light radius 2 on the worldmap.
  if redo_lite then self.player:recalcEverything() end

  if leaving_worldmap and not self.player:enoughEnergy() then
    -- Dodgy Hack(TM):  If we leave the worldmap in the middle of spending
    -- our increased movement energy, refund enough for the player to act
    -- first.
    self.player:useEnergy(self.player.energy.value - self.energy_to_act)
  end

  -- Remove temporary effects that are flagged as 'cancel_on_level_change'.
  local effs = {}
  for eff_id, p in pairs(self.player.tmp) do
    local def = self.player:getEffectFromId(eff_id)
    if def.cancel_on_level_change then
      effs[#effs+1] = eff_id
      if type(def.cancel_on_level_change) == 'function' then
	def.cancel_on_level_change(self.player, p)
      end
    end
  end
  for _, eff_id in ipairs(effs) do
    self.player:removeEffect(eff_id)
  end

  -- Apply zone effects.  T4 applies these to all actors, but T2 doesn't
  -- really support zone-wide effects for anyone but the player.
  if self.level.data.effects then
    for _, eff_id in ipairs(self.level.data.effects) do
      self.player:setEffect(eff_id, 1, {})
    end
  end

  class:triggerHook {
    'Game:enterLevel',
    zone = self.zone,
    level = self.level,
    lev = lev,
    old_zone = old_zone,
    old_lev = old_lev,
    params = params,
  }
  self.player:fireCallbacks('on_enter_level', self.zone, self.level, lev, old_zone, old_lev, params)

  -- Do the following in dungeons only.
  if util.zone_is(self.zone, 'dungeon') then
    -- Check fates.
    self.player:checkFates(self.zone, self.level)

    class:triggerHook {
      'Game:levelFeeling',
      zone = self.zone,
      level = self.level,
      lev = lev,
    }

    -- Update recall depth.
    self.state:updateRecallDepth()

    if not game.state.birth.preserve and self.level.special_feeling and not self.level.feeling then
      -- "Special feeling" in non-preserve mode if an artifact is on the
      -- level.
      self.level.feeling = 1
    end
    if self.level.rating and not self.level.feeling then
      -- Convert the computed level rating into a level feeling index.
      self.level.feeling = self.level.rating > 100 and 2 or
			   self.level.rating > 80 and 3 or
			   self.level.rating > 60 and 4 or
			   self.level.rating > 40 and 5 or
			   self.level.rating > 30 and 6 or
			   self.level.rating > 20 and 7 or
			   self.level.rating > 10 and 8 or
			   self.level.rating > 0 and 9 or 10
      -- TODO feeling = 0 if we weren't on the previous level long enough.
      self.level.rating = nil
      self.level.special_feeling = nil
    end

    local feeling = self.level.data.enter_desc
    if not feeling and self.level.feeling then
      self.level.feeling = util.minBound(self.level.feeling, 0, 10)
      feeling = level_feelings[self.level.feeling]
    end
    if feeling then game.log(feeling) end
  end
end

function _M:getPlayer()
  return self.player
end

--- Says if this savefile is usable or not
function _M:isLoadable()
  return not self:getPlayer(true).dead
end

function _M:tick()
  if self.level then
    self:targetOnTick()

    engine.GameTurnBased.tick(self)
    -- Fun stuff: this can make the game realtime, although calling it in display() will make it work better
    -- (since display is on a set FPS while tick() ticks as much as possible
    -- engine.GameEnergyBased.tick(self)
  end
  -- When paused (waiting for player input) we return true: this means we wont be called again until an event wakes us
  if self.paused and not savefile_pipe.saving then return true end
end

--- Called every game turn
function _M:onTurn()
  local plyr_turn = (self.turn % 10 == 0)

  -- Process overlay effects every game turn, but only do damage every
  -- normal-speed player turn.
  self.level.map:processEffects(not plyr_turn)

  if plyr_turn then	-- Once every normal-speed player turn
    -- Respawn monsters.
    local respawn_in = self.level.data.generator.actor and self.level.data.generator.actor.respawn_in
    if respawn_in and rng.chance(respawn_in) then
      -- Respawn monsters.
      local gen = self.zone:getGenerator('actor', self.level, self.level.data.generator.actor)
      gen:generateOne{ distance_from_player=25 }
    end

    -- TODO Day/night cycle.

    if not self.day_of_year or self.day_of_year ~= self.calendar:getDayOfYear(self.turn) then
      -- Log the calendar day.
      self.log(self.calendar:getTimeDate(self.turn))
      self.day_of_year = self.calendar:getDayOfYear(self.turn)
    end
  end
end

function _M:addParticle(x, y, name, args)
  local Particles = require 'engine.Particles'
  local p = Particles.new(name, 1, args)
  self.particles[p] = {x=x, y=y}
end

function _M:display(nb_keyframe)
  -- If switching resolution, blank everything but the dialog
  if self.change_res_dialog then
    engine.GameTurnBased.display(self, nb_keyframe)
    return
  end

  if self.full_fbo_shader and self.full_fbo_shader.shad then
    self.full_fbo_shader.shad:uniGamma(config.settings.gamma_correction/100)
  end

  if self.full_fbo then self.full_fbo:use(true) end

  -- Now the map, if any
  -- [Loosely based on T4's Game:displayMap()]
  if self.level and self.level.map and self.level.map.finished then
    -- Display the map and compute FOV for the player if needed
    local map = self.level.map
    if map.changed then self.player:playerFOV() end

    if self.fbo then
      -- Display using framebuffer with shaders
      self.fbo:use(true)
      map:display(0, 0, nb_keyframe, nil, self.fbo)
      self.fbo:use(false, self.full_fbo)
      -- Second pass to apply distorting effects
      self.fbo2:use(true)
      self.fbo:toScreen(0, 0, map.viewport.width, map.viewport.height)
      core.particles.drawAlterings()
      self.fbo2:use(false, self.full_fbo)
      _2DNoise:bind(1, false)
      self.fbo2:postEffects(self.fbo, self.full_fbo, map.display_x, map.display_y, map.viewport.width, map.viewport.height, unpack(self.posteffects_use))
      -- Display the targetting system if active
      self.target:display(nil, nil, self.full_fbo, nb_keyframes)
    else
      -- Basic display with no FBOs
      self.level.map:display(nil, nil, nb_keyframe)
      -- Display the targetting system if active
      self.target:display(nil, nil, self.full_fbo, nb_keyframes)
      core.particles.drawAlterings()
    end

    -- And the minimap
    self.level.map:minimapDisplay(self.w - 200, 20, util.bound(self.player.x - 25, 0, self.level.map.w - 50), util.bound(self.player.y - 25, 0, self.level.map.h - 50), 50, 50, 0.6)
  end

  -- We display the player's interface
  self.flash:toScreen(nb_keyframe)
  self.logdisplay:toScreen()
  self.stats_display:toScreen()
  if self.show_npc_list then
    self.npcs_display:toScreen()
  else
    self.hotkeys_display:toScreen()
  end

  -- Particles go above the hotkey toolbar, since for now they're just the
  -- hotkey feedback particles.
  if next(self.particles) then
    for p, pos in pairs(self.particles) do
      if p.ps:isAlive() then
	p.ps:toScreen(pos.x, pos.y, true, 1)
      else
	self.particles[p] = nil
      end
    end
  end

  -- "Big News"
  self.bignews:display(nb_keyframe)

  if core.key.modState('caps') and self.level and self.level.map and self.level.map.finished then
    -- Scroll mode; inform the player.
    local w = self.level.map.viewport.width/2
    local h = w * self.caps_scroll.h/self.caps_scroll.w
    local x = self.level.map.display_x + (self.level.map.viewport.width - w)/2
    local y = self.level.map.display_y + (self.level.map.viewport.height - w)/2
    local tex_w = self.caps_scroll[2] * w/self.caps_scroll.w
    local tex_h = self.caps_scroll[3] * h/self.caps_scroll.h
    self.caps_scroll[1]:toScreenFull(x, y, w, h, tex_w, tex_h, 1, 1, 1, 0.5)
  end

  if self.player then self.player.changed = false end

  engine.GameTurnBased.display(self, nb_keyframe)

  -- Tooltip is displayed over all else.  Put it on the SE corner unless
  -- current mouse coords is in the way.
  local mx, my, button = core.mouse.get()
  local tip = self.tooltip
  if tip.w and mx > self.w - tip.w and my > Tooltip:tooltip_bound_y2() - tip.h and not tip.locked then
    self:targetDisplayTooltip(Map.display_x, self.h, false, nb_keyframe)
  else
    self:targetDisplayTooltip(self.w, self.h, false, nb_keyframe)
  end

  if self.full_fbo then
    self.full_fbo:use(false)
    self.full_fbo:toScreen(0, 0, self.w, self.h, self.full_fbo_shader.shad)
  end
end

local all_colors = {
  colors.C_d, colors.C_w, colors.C_s, colors.C_o,
  colors.C_r, colors.C_g, colors.C_b, colors.C_u,
  colors.C_D, colors.C_W, colors.C_v, colors.C_y,
  colors.C_R, colors.C_G, colors.C_B, colors.C_U,
}
local rand_mon = {
  'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
  'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', 'A', 'B',
  'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
  'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 
}
local rand_obj = {
  '?', '/', '|', '\\', '"', '!', '$', '(', ')', '_', '-', '=', '[', ']',
  '{', '}', ',', '~', 
}
local hallu_mon_cache = {}
local hallu_obj_cache = {}

function _M:tweakVisuals(t)
  if not self.level.map.actor_player then return end
  t = t or {}
  local viewer = self.level.map.actor_player
  local hallu = viewer:has('HALLUCINATION')
  local ascii = config.settings.tome2.tiles == 'ascii'
  local tiles = config.settings.tome2.tiles

  if not ascii then
    if #hallu_mon_cache == 0 then
      for _, fn in ipairs(fs.list('/data/gfx/'..tiles..'/npc/')) do
	if fn:find('.png$') then table.insert(hallu_mon_cache, 'npc/'..fn) end
      end
    end
    if #hallu_obj_cache == 0 then
      for _, fn in ipairs(fs.list('/data/gfx/'..tiles..'/object/')) do
	if fn:find('.png$') then table.insert(hallu_obj_cache, 'object/'..fn) end
      end
    end
  end

  local function shimmer_rd(e)
    local tt = { display=e.display, image=e.image }
    local ff = e.flags or {}
    if ff.ATTR_MULTI then
      -- Pick a random color, entity-specific if so defined.
      tt.color = rng.table(e.multi_colors or all_colors)
    end
    if ff.ATTR_CLEAR and e.x and e.y then
      -- Use color of underlying grid.
      local g = self.level.map(e.x, e.y, Map.TERRAIN)
      tt.color = { r=g.color_r, g=g.color_g, b=g.color_b }
    end
    if ff.CHAR_CLEAR and e.x and e.y then
      -- Use char of underlying grid.
      local g = self.level.map(e.x, e.y, Map.TERRAIN)
      tt.display = e.display
      tt.image = e.image
    end
    if e:getEntityKind() == 'actor' then
      return Actor.new(tt)
    elseif e:getEntityKind() == 'object' then
      return Object.new(tt)
    end
  end

  local function hallu_rd(e, disps, images)
    local tt = { color=rng.table(all_colors) }
    if e:getEntityKind() == 'actor' then
      tt.display, tt.image = rng.table(rand_mon), rng.table(hallu_mon_cache)
      return Actor.new(tt)
    elseif e:getEntityKind() == 'object' then
      tt.display, tt.image = rng.table(rand_obj), rng.table(hallu_obj_cache)
      return Object.new(tt)
    end
  end

  -- Shimmer and hallucinate all actors we can see by any means.
  for _, e in pairs(game.level.entities) do
    if e:getEntityKind() == 'actor' then
      if t.clear_hallu then
	-- Aggressively clear all actors.
	e:replaceDisplay('hallucination', nil)
	if e:has('MIMIC') then
	  -- Questionable Hack(TM):  Also aggressively clear mimics'
	  -- replace_display object.
	  local rd_o = e.replace_displays.mimic
	  if rd_o and rd_o:getEntityKind() == 'object' then
	    rd_o:replaceDisplay('hallucination', nil)
	  end
	end
      end
      local los = viewer.fov.actors[e] and true or false
      if not e.player and viewer:canSee(e, los, (los and 100 or 0)) then
	if hallu then e:replaceDisplay('hallucination', hallu_rd(e)) end
	if e:hasTweakableVisuals() then
	  e:replaceDisplay('shimmer', shimmer_rd(e))
	end
      end
    end
  end

  if t.clear_hallu then
    -- Aggressively clear all objects on the floor.
    for x = 0, self.level.map.w - 1 do
      for y = 0, self.level.map.h - 1 do
	local changed = false
	local function cb(o, x, y, i)
	  changed = o:replaceDisplay('hallucination', nil) or changed
	end
	self.level.map:objectsApply(x, y, cb)
	if changed then self.level.map:updateMap(x, y) end
      end
    end
  end
  -- Shimmer and hallucinate objects in LOS on seen grids.
  local function cb1(_, x, y) return self.level.map:opaque(x, y) end
  local function cb2(_, x, y)
    if self.level.map.seens(x, y) then
      local o = self.level.map:getObject(x, y, 1)
      -- Questionable Hack(TM):  If a mimic is present on the grid,
      -- hallucinate/shimmer its replace_display object instead of the
      -- floor object.
      local a = self.level.map(x, y, Map.ACTOR)
      if a and a:has('MIMIC') then
	local rd_o = a.replace_displays.mimic
	if rd_o and rd_o:getEntityKind() == 'object' then o = rd_o end
      end
      if o then
	local changed = false
	if hallu then
	  changed = o:replaceDisplay('hallucination', hallu_rd(o)) or changed
	end
	if o.flags and o.flags.ATTR_MULTI then
	  changed = o:replaceDisplay('shimmer', shimmer_rd(o)) or changed
	end
	if changed then self.level.map:updateMap(x, y) end
      end
    end
  end
  core.fov.calc_circle(self.player.x, self.player.y, self.level.map.w, self.level.map.h, self.player.sight or 20, cb1, cb2, nil)
end

--- Setup the keybinds
function _M:setupCommands()
  -- Make targeting work
  self.normal_key = self.key
  self:targetSetupKey()

  local function move_or_scroll(dir)
    if core.key.modState('caps') and self.level then
      self.level.map:scrollDir(dir)
    else
      self.player:moveDir(dir)
    end
  end
  local function tunnel_dir(dir)
    local p = self.player
    if dir == 5 or not p then return end
    local dx, dy = util.dirToCoord(dir, p.x, p.y)
    p:useTalent(p.T_OBJ_DIG, nil, nil, nil, { x=p.x+dx, y=p.y+dy })
  end

  local function do_in_coroutine(cb)
    local co = coroutine.create(cb)
    local ok, err = coroutine.resume(co)
    if not ok then
      print(debug.traceback(co))
      if err then error(err) end
    end
  end

  -- One key handled for normal function
  self.key:unicodeInput(true)
  self.key:addBinds
  {
    -- Movements
    MOVE_LEFT = function() move_or_scroll(4) end,
    MOVE_RIGHT = function() move_or_scroll(6) end,
    MOVE_UP = function() move_or_scroll(8) end,
    MOVE_DOWN = function() move_or_scroll(2) end,
    MOVE_LEFT_UP = function() move_or_scroll(7) end,
    MOVE_LEFT_DOWN = function() move_or_scroll(1) end,
    MOVE_RIGHT_UP = function() move_or_scroll(9) end,
    MOVE_RIGHT_DOWN = function() move_or_scroll(3) end,
    MOVE_STAY = function() self.player:stay() end,

    RUN_LEFT = function() self.player:runInit(4) end,
    RUN_RIGHT = function() self.player:runInit(6) end,
    RUN_UP = function() self.player:runInit(8) end,
    RUN_DOWN = function() self.player:runInit(2) end,
    RUN_LEFT_UP = function() self.player:runInit(7) end,
    RUN_LEFT_DOWN = function() self.player:runInit(1) end,
    RUN_RIGHT_UP = function() self.player:runInit(9) end,
    RUN_RIGHT_DOWN = function() self.player:runInit(3) end,

    RUN_AUTO = function()
      if self.level and self.zone then
	local seen = {}
	-- Check for visible monsters.  Only see LOS actors, so telepathy
	-- wont prevent it
	local function cb1(_, x, y) return self.level.map:opaque(x, y) end
	local function cb2(_, x, y)
	  local actor = self.level.map(x, y, self.level.map.ACTOR)
	  if actor and actor ~= self.player and
	     self.player:reactionToward(actor) < 0 and
	     self.player:canSee(actor) and self.level.map.seens(x, y)
	  then
	    seen[#seen + 1] = {x=x, y=y, actor=actor}
	  end
	end
	core.fov.calc_circle(self.player.x, self.player.y,
			     self.level.map.w, self.level.map.h,
			     self.player.sight or 20, cb1, cb2, nil)
	if self.zone.no_autoexplore or self.level.no_autoexplore then
	  self.log("You may not auto-explore this level.")
	elseif #seen > 0 then
	  local dir = game.level.map:compassDirection(seen[1].x - self.player.x, seen[1].y - self.player.y)
	  self.log("You may not auto-explore with enemies in sight (%s to the %s%s)!", seen[1].actor.name, dir, self.level.map:isOnScreen(seen[1].x, seen[1].y) and "" or " - offscreen")
	  for _, node in ipairs(seen) do
	    node.actor:addParticles(engine.Particles.new("notice_enemy", 1))
	  end
	elseif not self.player:autoExplore() then
	  self.log("There is nowhere left to explore.")
	end
      end
    end,

    TUNNEL_LEFT = function() tunnel_dir(4) end,
    TUNNEL_RIGHT = function() tunnel_dir(6) end,
    TUNNEL_UP = function() tunnel_dir(8) end,
    TUNNEL_DOWN = function() tunnel_dir(2) end,
    TUNNEL_LEFT_UP = function() tunnel_dir(7) end,
    TUNNEL_LEFT_DOWN = function() tunnel_dir(1) end,
    TUNNEL_RIGHT_UP = function() tunnel_dir(9) end,
    TUNNEL_RIGHT_DOWN = function() tunnel_dir(3) end,

    SHOW_INVENTORY = function()
      if self.player.no_inventory_access then return end
      local d
      local function use_cb(_, _, _, _, stop)
	d:generate()
	d:generateList()
	if stop then self:unregisterDialog(d) end
      end
      local function inven_cb(o, inven, floor, item, btn, event)
	local ud = UseItemDialog.new(event == 'button', self.player, o, item, inven, floor, use_cb)
	self:registerDialog(ud)
      end
      d = self.player:showEquipInvenFloor(self.player:weightDynamicTitle('Inventory'), nil, inven_cb, nil, {allow_drag=true})
    end,
    SHOW_EQUIPMENT =  'SHOW_INVENTORY',
    WEAR_ITEM = function()
      if self.player.no_inventory_access then return end
      self.player:playerWear()
    end,
    TAKEOFF_ITEM = function()
      if self.player.no_inventory_access then return end
      self.player:playerTakeoff()
    end,
    USE_ITEM = function()
      if self.player.no_inventory_access then return end
      self.player:playerUseItem()
    end,

    OPEN = function()
      -- TODO Check whether player can actually open stuff.
      do_in_coroutine(function() self.player:playerOpen() end)
    end,
    CLOSE = function()
      -- TODO Check whether player can actually close doors.
      do_in_coroutine(function() self.player:playerCloseDoor() end)
    end,
    BASH_DOOR = function()
      -- TODO Check whether player can actually bash doors.
      do_in_coroutine(function() self.player:playerBashDoor() end)
    end,

    PICKUP_FLOOR = function()
      if not self.player.no_inventory_access then
	self.player:playerPickup()
      end
    end,
    DROP_FLOOR = function()
      if not self.player.no_inventory_access then
	self.player:playerDrop()
      end
    end,

    DISARM = function()
      -- TODO Is there any condition under which we wouldn't be allowed to
      -- do this?  Body with no arms?
      do_in_coroutine(function() self.player:playerDisarm() end)
    end,
    FORCE_TRIGGER = function()
      do_in_coroutine(function() self.player:playerDisarm(true) end)
    end,

    SHOW_QUESTS = function ()
      self:registerDialog(require('mod.dialogs.ShowQuests').new(self.player))
    end,

    -- Hotkeys
    HOTKEY_1 = function() self.player:activateHotkey(1) end,
    HOTKEY_2 = function() self.player:activateHotkey(2) end,
    HOTKEY_3 = function() self.player:activateHotkey(3) end,
    HOTKEY_4 = function() self.player:activateHotkey(4) end,
    HOTKEY_5 = function() self.player:activateHotkey(5) end,
    HOTKEY_6 = function() self.player:activateHotkey(6) end,
    HOTKEY_7 = function() self.player:activateHotkey(7) end,
    HOTKEY_8 = function() self.player:activateHotkey(8) end,
    HOTKEY_9 = function() self.player:activateHotkey(9) end,
    HOTKEY_10 = function() self.player:activateHotkey(10) end,
    HOTKEY_11 = function() self.player:activateHotkey(11) end,
    HOTKEY_12 = function() self.player:activateHotkey(12) end,
    HOTKEY_SECOND_1 = function() self.player:activateHotkey(13) end,
    HOTKEY_SECOND_2 = function() self.player:activateHotkey(14) end,
    HOTKEY_SECOND_3 = function() self.player:activateHotkey(15) end,
    HOTKEY_SECOND_4 = function() self.player:activateHotkey(16) end,
    HOTKEY_SECOND_5 = function() self.player:activateHotkey(17) end,
    HOTKEY_SECOND_6 = function() self.player:activateHotkey(18) end,
    HOTKEY_SECOND_7 = function() self.player:activateHotkey(19) end,
    HOTKEY_SECOND_8 = function() self.player:activateHotkey(20) end,
    HOTKEY_SECOND_9 = function() self.player:activateHotkey(21) end,
    HOTKEY_SECOND_10 = function() self.player:activateHotkey(22) end,
    HOTKEY_SECOND_11 = function() self.player:activateHotkey(23) end,
    HOTKEY_SECOND_12 = function() self.player:activateHotkey(24) end,
    HOTKEY_THIRD_1 = function() self.player:activateHotkey(25) end,
    HOTKEY_THIRD_2 = function() self.player:activateHotkey(26) end,
    HOTKEY_THIRD_3 = function() self.player:activateHotkey(27) end,
    HOTKEY_THIRD_4 = function() self.player:activateHotkey(28) end,
    HOTKEY_THIRD_5 = function() self.player:activateHotkey(29) end,
    HOTKEY_THIRD_6 = function() self.player:activateHotkey(30) end,
    HOTKEY_THIRD_7 = function() self.player:activateHotkey(31) end,
    HOTKEY_THIRD_8 = function() self.player:activateHotkey(32) end,
    HOTKEY_THIRD_9 = function() self.player:activateHotkey(33) end,
    HOTKEY_THIRD_10 = function() self.player:activateHotkey(34) end,
    HOTKEY_THIRD_11 = function() self.player:activateHotkey(35) end,
    HOTKEY_THIRD_12 = function() self.player:activateHotkey(36) end,
    HOTKEY_PREV_PAGE = function()
      self.player:prevHotkeyPage()
      self.log('Hotkey page %d is now displayed.', self.player.hotkey_page)
    end,
    HOTKEY_NEXT_PAGE = function()
      self.player:nextHotkeyPage()
      self.log('Hotkey page %d is now displayed.', self.player.hotkey_page)
    end,

    -- Actions
    CHANGE_LEVEL = function(sym, ctrl, shift, alt, meta, unicode, isup, key)
      -- Timed Roots blocks this.
      -- TODO also prevent if player is controlling a monster
      if self.player:hasEffect(self.player.EFF_TREE_ROOTS) then return end
      -- Most places we allow either < or > to change level, a la T4, but
      -- there are a few places that we care which way the player wants to
      -- go.
      local up = (unicode == '<')
      local e = self.level.map(self.player.x, self.player.y, Map.TERRAIN)
      if (util.zone_is(self.zone, 'wilderness') or util.zone_is(self.zone, 'town')) and up and self:okGoToWorldMap() then
	-- Up to the world map.  We check this first in case player is
	-- standing on dungeon entrance in wilderness zone (or quest zone
	-- entrance in town).
	self:changeLevel(1, 'worldmap')
      elseif self.player:enoughEnergy() and e.change_level then
	if e.ok_change_level and not e:ok_change_level(self.player) then return end
	self:changeLevel(e.change_zone and e.change_level or self.level.level + e.change_level, e.change_zone, e.change_zone_params)
      elseif self.player:enoughEnergy() and e.use_jumpgate then
	e:use_jumpgate(self.player)
      elseif self.zone.worldmap and not up then
	-- Down from the worldmap into the wilderness; anything that would
	-- enter us into a zone was caught above (I think...).
	self:changeLevel(1, 'wilderness')
      else
	self.log("There is no way out of this level here.")
      end
    end,

    REST = function()
      self.player:restInit()
    end,
    REST_TIMED = function()
      local f = function(qty)
	if qty > 1 then game.player:restInit(qty - 1) end
      end
      local GetQuantity = require 'engine.dialogs.GetQuantity'
      game:registerDialog(GetQuantity.new('Timed Rest', 'Turns:', 0, 999, f))
    end,

    SEARCH = function()
      self.player:playerSearch()
    end,

    USE_TALENTS = function()
      local UseTalents = require 'mod.dialogs.UseTalents'
      self:registerDialog(UseTalents.new(self.player))
    end,
    USE_THAUMATURGY = function()
      local UseThaumaturgy = require 'mod.dialogs.UseThaumaturgy'
      self:registerDialog(UseThaumaturgy.new(self.player))
    end,

    SAVE_GAME = function()
      self:saveGame()
    end,

    SHOW_CHARACTER_SHEET = function()
      self:registerDialog(require("mod.dialogs.CharacterSheet").new(self.player))
    end,

    SHOW_MESSAGE_LOG = function()
      --local ShowLog = require 'engine.dialogs.ShowLog'
      --self:registerDialog(ShowLog.new('Message Log', 0.6, self.logdisplay))
      self.logdisplay:showLogDialog('Message Log', 0.6)
    end,

    CENTER_ON_PLAYER = function()
      self.level.map:centerViewAround(self.player.x, self.player.y)
    end,

    -- Show time
    SHOW_TIME = function()
      self.log(self.calendar:getTimeDate(self.turn))
    end,

    -- Exit the game
    QUIT_GAME = function()
      self:onQuit()
    end,

    SCREENSHOT = function() self:saveScreenshot() end,

    EXIT = function()
      local GameMenu = require 'engine.dialogs.GameMenu'
      local menu
      local l = {
	"resume",
	{
	  'Game Options',
	  function()
	    self:unregisterDialog(menu)
	    local GameOptions = require 'mod.dialogs.GameOptions'
	    self:registerDialog(GameOptions.new())
	  end
	},
	{
	  'Change Melee Style',
	  function ()
	    self:unregisterDialog(menu)
	    self.player:selectMeleeStyle()
	  end
	},
	{
	  'Change Tactic/Explor',
	  function()
	    self:unregisterDialog(menu)
	    local TacticExplorDialog = require 'mod.dialogs.TacticExplorDialog'
	    self:registerDialog(TacticExplorDialog.new(self.player))
	  end
	},
	{
	  'Show Achievements',
	  function() 
	    self:unregisterDialog(menu)
	    local ShowAchievements = require 'mod.dialogs.ShowAchievements'
	    self:registerDialog(ShowAchievements.new('Tales of Middle Earth Achievements', self.player))
	  end
	},
	{
	  'Show Lore',
	  function() 
	    self:unregisterDialog(menu)
	    local ShowLore = require 'mod.dialogs.ShowLore'
	    self:registerDialog(ShowLore.new('Tales of Middle Earth Lore', self.player))
	  end
	},
	{
	  'Automatizer Rules',
	  function()
	    self:unregisterDialog(menu)
	    local AutomatizerRules = require 'mod.dialogs.AutomatizerRules'
	    self:registerDialog(AutomatizerRules.new())
	  end
	},
	{
	  'Show Recall Targets',
	  function()
	    self:unregisterDialog(menu)
	    if game.state.recall_town then
	      game.log('#LIGHT_BLUE#Town recall target:  ' .. game.state.recall_town.name .. '#LAST#')
	    end
	    if game.state.recall_dungeon then
	      game.log('#LIGHT_BLUE#Dungeon recall target:  ' .. game.state.recall_dungeon.name .. '#LAST#')
	    end
	  end
	},
	{
	  'View Fates',
	  function()
	    self:unregisterDialog(menu)
	    self.player:showFates()
	  end
	},
	"keybinds",
	"video",
	{
	  "Credits",
	  function()
	    self:unregisterDialog(menu)
	    local CreditsDialog = require 'mod.dialogs.CreditsDialog'
	    self:registerDialog(CreditsDialog.new())
	  end
	},
	"save",
	"quit",
	"exit"
      }
      if game.player:getSkillLevel('ALCHEMY') > 0 then
	local h = {
	  'Show Alchemy Recipes',
	  function()
	    self:unregisterDialog(menu)
	    local AlchemyRecipesDialog = require 'mod.dialogs.AlchemyRecipesDialog'
	    self:registerDialog(AlchemyRecipesDialog.new(game.player))
	  end,
	}
	-- Insert after "Show Recall Targets"
	table.insert(l, 7, h)
      end
      self:triggerHook {
	'Game:alterGameMenu',
	menu = l,
	unregister = function() self:unregisterDialog(menu) end,
      }
      menu = GameMenu.new(l)
      self:registerDialog(menu)
    end,

    -- Lua console
    LUA_CONSOLE = function()
      self:registerDialog(DebugConsole.new())
    end,
    -- Debug dialog
    DEBUG_MODE = function()
      self:registerDialog(require('mod.dialogs.debug.DebugDialog').new())
    end,

    -- Toggle monster list
    TOGGLE_NPC_LIST = function()
      self.show_npc_list = not self.show_npc_list
      self.player.changed = true
    end,

    TACTICAL_DISPLAY = function()
      if Map.view_faction then
	self.always_target = nil
	Map:setViewerFaction(nil)
      else
	self.always_target = true
	Map:setViewerFaction("players")
      end
    end,

    LOOK_AROUND = function()
      self.flash:empty(true)
      self.flash(self.flash.GOOD, "Looking around... (direction keys to select interesting things, shift+direction keys to move freely)")
      local co = coroutine.create(function() self.player:getTarget{type="hit", no_restrict=true, range=2000} end)
      local ok, err = coroutine.resume(co)
      if not ok and err then print(debug.traceback(co)) error(err) end
    end,

    LEVELUP = function()
      local SkillDialog = require 'mod.dialogs.SkillDialog'
      game:registerDialog(SkillDialog.new(self.player, nil))
    end,
  }
  self.key:setCurrent()
end

-- Return true if we can go to the world map.  The caller has already
-- confirmed that we're in the wilderness or a town.
-- [cf. src/dungeon.c lines 3776-3814]
function _M:okGoToWorldMap(silent)
  if self.player.food < util.FOOD.hungry then
    if not silent then self.log('You are too hungry to travel.') end
    return false
  end
  -- TODO Vampires can't travel during the day.
  for _, o in ipairs(self.player:getInven('LITE')) do
    if o.fuel_lite then
      if not silent then
	self.log('You should take off your %s before traveling.', o:getName{no_count=true})
      end
      return false
    end
  end
  if self.player:has('CUT') or self.player:has('POISON') then
    if not silent then self.log('You are too injured to travel.') end
    return false
  end
  if self.player:has('WILD_AMBUSH') then
    if not silent then
      self.log('To flee the ambush you must reach the edge of the map.')
    end
    return false
  end
  -- TODO Allow veto via hook or actor callback.
  -- [That would involve wrapping this whole thing in a coroutine, though,
  --  so that the hooks/callbacks could use dialogs.]

  -- No problems; we're good.
  return true
end

function _M:setupMouse(reset)
  if reset then self.mouse:reset() end

  self.mouse:registerZone(Map.display_x, Map.display_y, Map.viewport.width, Map.viewport.height, function(button, mx, my, xrel, yrel, bx, by, event, extra)
    -- Handle targeting
    if self:targetMouse(button, mx, my, xrel, yrel, event) then return end

    -- Handle right-click menu
    if button == 'right' and not xrel and not yrel then
      self:mouseRightClick(mx, my, extra)
    end

    -- Handle the mouse movement/scrolling
    self.player:mouseHandleDefault(self.key, self.key == self.normal_key, button, mx, my, xrel, yrel, event)
  end)

  -- Scroll message log
  self.mouse:registerZone(self.logdisplay.display_x, self.logdisplay.display_y, self.logdisplay.w, self.logdisplay.h, function(button)
    if button == "wheelup" then self.logdisplay:scrollUp(1) end
    if button == "wheeldown" then self.logdisplay:scrollUp(-1) end
  end, {button=true})

  -- Use hotkeys with mouse
  self.mouse:registerZone(self.hotkeys_display.display_x, self.hotkeys_display.display_y, self.hotkeys_display.w, self.hotkeys_display.h, function(button, mx, my, xrel, yrel, bx, by, event)
    self.hotkeys_display:onMouse(button, mx, my, event == "button", function(text) self:tooltipDisplayAtMap(self.w, self.h, tostring(text)) end)
  end)

  -- Tooltips for effects/sustains in the stats display
  self.mouse:registerZone(self.stats_display.display_x, self.stats_display.display_y, self.stats_display.w, self.stats_display.h, function(button, mx, my, xrel, yrel, bx, by, event)
    self.stats_display:onMouse(button, mx, my, event == "button", function(text) self:tooltipDisplayAtMap(self.w, self.h, tostring(text)) end)
  end)

  self.mouse:setCurrent()
end

function _M:mouseRightClick(mx, my, extra)
  if not self.level then return end
  local tmx, tmy = self.level.map:getMouseTile(mx, my)
  local MapMenu = require 'mod.dialogs.MapMenu'
  self:registerDialog(MapMenu.new(mx, my, tmx, tmy, extra and extra.add_map_action))
end

--- Ask if we really want to close, if so, save the game first
function _M:onQuit()
  self.player:runStop('quitting')
  self.player:restStop('quitting')

  if not self.quit_dialog and not self.player.dead and not self:hasDialogUp() then
    local cb = function(ok)
      if ok then
	savefile_pipe:push(game.save_name, "game", game)
	savefile_pipe:forceWait()
	util.showMainMenu()
      end
      self.quit_dialog = nil
    end
    self.quit_dialog = Dialog:yesnoPopup('Save', 'Save and return to main menu?', cb)
  end
end

-- Our expanded QuitDialog has options to save/abandon character and
-- continue playing/return to main menu/quit.
function _M:onExit()
  self.player:runStop('quitting')
  self.player:restStop('quitting')

  if not self.quit_dialog and not self.player.dead and not self:hasDialogUp() then
    self.quit_dialog = QuitDialog.new()
    self:registerDialog(self.quit_dialog)
  end
end

--- Requests the game to save
function _M:saveGame()
  -- savefile_pipe is created as a global by the engine
  savefile_pipe:push(self.save_name, "game", self)
  world:saveWorld()
  self.log("Saving game...")
end

function _M:getStore(name, type)
  -- Return the persisted store, if any.
  local lev = self.zone:level_adjust_level(self.level, 'trap')
  local key = self.zone.short_name .. '-' .. lev .. '/' .. type .. '-' .. name
  self.state.stores = self.state.stores or {}
  if self.state.stores[key] then return self.state.stores[key] end

  -- Nope, create a new one.
  local e
  if type == 'home' then
    e = Home.homes_def[name]:clone()
  elseif type == 'museum' then
    e = MathomHouse.museums_def[name]:clone()
  else
    e = Store.stores_def[name]:clone()
  end
  -- We have to do this to resolve store owners.
  e:resolve()

  -- Save the store.
  self.state.stores[key] = e
  return e
end

-- Returns all stores of the specified type (if any) on the current level.
function _M:getStores(type)
  local lev = self.zone:level_adjust_level(self.level, 'trap')
  local key_pfx = '^' .. self.zone.short_name .. '-' .. lev .. '/'
  if type then key_pfx = key_pfx .. type .. '-' end
  local ret = {}
  for key, store in pairs(self.state.stores or {}) do
    if key:find(key_pfx) then ret[#ret+1] = store end
  end
  return ret
end

function _M:getStoreOwner(name)
  return StoreOwner.owners_def[name]:clone()
end

function _M:onRegisterDialog(d)
  self.tooltip_x, self.tooltip_y = nil, nil
  if self.player then self.player:updateMainShader() end
end

function _M:onUnregisterDialog(d)
  self.tooltip_x, self.tooltip_y = nil, nil
  if self.player then
    self.player:updateMainShader()
    self.player.changed = true
  end
end
