-- $Id: Static.lua 3850 2020-04-05 20:13:34Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2020 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Super = require 'engine.generator.map.Static'
local Map = require "mod.class.Map"

module(..., package.seeall, class.inherit(Super))

function _M:getLoader(t)
  local env = Super.getLoader(self, t)
  env.include = function(file)
    local g = self:getLoader(t)
    local ret, err = self:loadLuaInEnv(g, file)
    if not ret and err then error(err) end
    return ret
  end
  return env
end

-- Duplicated and modified from Super:resolve() to support the use of a
-- "chained" resolver to handle grids like 'outer_wall'.
function _M:resolve(typ, c)
  local res
  if typ then
    if not self.tiles[c] or not self.tiles[c][typ] then return end
    res = self.tiles[c][typ]
  else
    if not self.tiles[c] then return end
    res = self.tiles[c]
  end

  -- Full-blown grids return as is.
  if type(res) == "table" and (res.__ATOMIC or res.__CLASSNAME) then
    return res
  end
  -- Resolve (so to speak) which grid we want.
  if type(res) == 'function' then
    res = res()
  elseif type(res) == 'table' then
    res = rng.table(1, #res)
  end
  -- If we have the grid, return it.
  if self.grid_list[res] then return self.grid_list[res] end
  -- Otherwise, if we have a chained resolver, try it.
  if self.chain_resolve then return self.chain_resolve(res) end
  -- This is probably not good...
  print(("[STATIC] unable to resolve grid type '%s'"):format(tostring(res)))
  return nil
end

function _M:roomMapAddEntity(i, j, type, e)
  Super.roomMapAddEntity(self, i, j, type, e)

  if type == 'object' then
    -- Mark object as having been found on the floor, unless we've already
    -- set found location information.
    if not e.found then
      e:setFoundHere({ type='floor' }, self.zone, self.level)
    end
  end
end

-- Duplicated and slightly modified from Super:generate(); we need a bit
-- more flexibility in specifying and modifying actors in defineTile().
function _M:generate(lev, old_lev)
  local function get_field(c, fld)
    if type(c) == 'string' then return self.tiles[c] and self.tiles[c][fld] end
    return c[fld] and self.tiles[c[fld]]
  end

  local status_hack = {
    use_wall_grid = 'secret_door',
    door_lock_power = 'lock_power',
    door_jam_power = 'jam_power',
  }

  for i = 1, self.gen_map.w do
    for j = 1, self.gen_map.h do
      local c = self.gen_map[i][j]
      local g = type(c) == 'string' and self:resolve('grid', c) or self:resolve(nil, c.grid)
      if g then
	-- Convoluted Hack(TM):  Pull some flags out of the status{} table
	-- for some grids, notably for locked/jammed/secret doors.
	local status = get_field(c, 'status')
	for k, v in pairs(status_hack) do spell[k] = status and status[v] end
	if g.force_clone then g = g:clone() end
	g:resolve()
	g:resolve(nil, true)
	self.map(i-1, j-1, Map.TERRAIN, g)
	g:check("addedToLevel", self.level, i-1, j-1)
	g:check("on_added", self.level, i-1, j-1)
	-- Clean these up behind ourself.
	for k, _ in ipairs(status_hack) do spell[k] = nil end
      end

      if self.status_all then
	local s = table.clone(self.status_all)
	if s.lite then
	  self.level.map.lites(i-1, j-1, true)
	  s.lite = nil
	end
	if s.remember then
	  self.level.map.remembers(i-1, j-1, true)
	  s.remember = nil
	end
	if s.special then
	  self.map.room_map[i-1][j-1].special = s.special
	  s.special = nil
	end
	if s.room_map then
	  for k, v in pairs(s.room_map) do self.map.room_map[i-1][j-1][k] = v end
	  s.room_map = nil
	end
	if pairs(s) then
	  for k, v in pairs(s) do self.level.map.attrs(i-1, j-1, k, v) end
	end
      end
    end
  end

  -- generate the rest after because they might need full map data to be correctly made
  for i = 1, self.gen_map.w do
    for j = 1, self.gen_map.h do
      local c = self.gen_map[i][j]
      local actor = get_field(c, 'actor')
      local trap = get_field(c, 'trap')
      local object = get_field(c, 'object')
      local trigger = get_field(c, 'trigger')
      local status = get_field(c, 'status')
      local define_spot = get_field(c, 'define_spot')

      -- Vaguely Convoluted Hack(TM):  Allow objects, actors and traps to
      -- specify a probability of being generated, and specify exclusions
      -- for other probability checks (this is mainly to implement the '*'
      -- and ',' map tile markers for vaults).  Note that the order of
      -- checking is object->trap->actor, even though the parameter order
      -- in defineTile is object->actor->trap.
      local placed = {}
      local function check_cross_prob(def, flag)
	if type(def) ~= 'table' or (not def.prob and not def.not_if) then
	  return def
	end
	if def.not_if and placed[def.not_if] then return nil end
	if def.prob and not rng.percent(def.prob) then return nil end
	placed[flag] = true
	return def[1]
      end

      if trigger then
	local t, mod
	if type(trigger) == "string" then
	  t = self.zone:makeEntityByName(self.level, "trigger", trigger)
	elseif type(trigger) == "table" and trigger.random_filter then
	  mod = trigger.entity_mod
	  t = self.zone:makeEntity(self.level, "terrain", trigger.random_filter, nil, true)
	else
	  t = self.zone:finishEntity(self.level, "terrain", trigger)
	end

	if t then
	  if mod then t = mod(t) end
	  self:roomMapAddEntity(i-1, j-1, "trigger", t)
	end
      end

      if object then
	local o, mod
	local Object = require 'mod.class.Object'
	object = check_cross_prob(object, 'object')
	if not object then
	  -- check_cross_prob() probability check failed; do nothing.
	elseif type(object) == "string" then
	  o = self.zone:makeEntityByName(self.level, "object", object)
	elseif type(object) == "table" and object.random_filter then
	  mod = object.entity_mod
	  if object.resolve_flags then
	    table.merge(Object.resolve_flags, object.resolve_flags, true)
	  end
	  local filter = self.zone:createWeightedObjectFilter(true, object.random_filter)
	  o = self.zone:makeEntity(self.level, "object", filter, nil, true)
	elseif type(object) == 'table' and object.name then
	  -- Mod to the parent method:  allow makeEntityByName() in
	  -- combination with specifying a modifier function and optional
	  -- resolve flags.
	  mod = object.entity_mod
	  if object.resolve_flags then
	    table.merge(Object.resolve_flags, object.resolve_flags, true)
	  end
	  o = self.zone:makeEntityByName(self.level, 'object', object.name)
	else
	  o = self.zone:finishEntity(self.level, "object", object)
	end

	if o then
	  if mod then o = mod(o) end
	  if status and status.obj_found then
	    -- Slight Hack(TM):  Borrow the status{} table to populate
	    -- found location information.
	    o:setFoundHere(status.obj_found, self.zone, self.level)
	  end
	  self:roomMapAddEntity(i-1, j-1, "object", o)
	end
      end

      if trap then
	local t, mod
	trap = check_cross_prob(trap, 'trap')
	if not trap then
	  -- check_cross_prob() probability check failed; do nothing.
	elseif type(trap) == "string" then
	  t = self.zone:makeEntityByName(self.level, "trap", trap)
	elseif type(trap) == "table" and trap.random_filter then
	  mod = trap.entity_mod
	  t = self.zone:makeEntity(self.level, "trap", trap.random_filter, nil, true)
	else
	  t = self.zone:finishEntity(self.level, "trap", trap)
	end
	if t then
	  if mod then t = mod(t) end
	  self:roomMapAddEntity(i-1, j-1, "trap", t)
	end
      end

      if actor then
	local m, mod
	local NPC = require 'mod.class.NPC'
	actor = check_cross_prob(actor, 'actor')
	if not actor then
	  -- check_cross_prob() probability check failed; do nothing.
	elseif type(actor) == "string" then
	  m = self.zone:makeEntityByName(self.level, "actor", actor)
	elseif type(actor) == "table" and actor.random_filter then
	  mod = actor.entity_mod
	  if actor.resolve_flags then
	    table.merge(NPC.resolve_flags, actor.resolve_flags, true)
	  end
	  m = self.zone:makeEntity(self.level, "actor", actor.random_filter, nil, true)
	elseif type(actor) == "table" and actor.name then
	  -- Mod to the parent method:  allow makeEntityByName() in
	  -- combination with specifying a modifier function.
	  mod = actor.entity_mod
	  if actor.resolve_flags then
	    table.merge(NPC.resolve_flags, actor.resolve_flags, true)
	  end
	  m = self.zone:makeEntityByName(self.level, "actor", actor.name, actor.force_unique)
	else
	  -- Slight Hack(TM):  Borrow the status{} table to convey the
	  -- 'force_unique' flag to Zone:finishEntity() as necessary.
	  -- quest_mon flag on the monster.
	  m = self.zone:finishEntity(self.level, "actor", actor, status and status.force_unique)
	end
	if m then
	  -- Slight Hack(TM):  Borrow the status{} table to install a
	  -- quest_mon flag on the monster.
	  if status and status.quest_mon then m.quest_mon = true end
	  if status and status.quest_on_die then
	    m.quest_on_die = status.quest_on_die
	  end
	  if mod then m = mod(m) end
	  self:roomMapAddEntity(i-1, j-1, "actor", m)
	end
      end

      if status then
	local s = table.clone(status)
	-- Slight Hack(TM), redux:  We don't want to put our assorted
	-- special-use flags in the grid status.
	s.quest_mon = nil
	s.quest_on_die = nil
	s.force_unique = nil
	s.obj_found = nil
	for _, v in pairs(status_hack) do s[v] = nil end
	if s.lite then
	  self.level.map.lites(i-1, j-1, true)
	  s.lite = nil
	end
	if s.remember then
	  self.level.map.remembers(i-1, j-1, true)
	  s.remember = nil
	end
	if s.special then
	  self.map.room_map[i-1][j-1].special = s.special
	  s.special = nil
	end
	if s.room_map then
	  for k, v in pairs(s.room_map) do self.map.room_map[i-1][j-1][k] = v end
	  s.room_map = nil
	end
	-- Questionable Hack(TM):  Allow a map of mod functions to modify
	-- status fields on installation.
	local mods = s.mods or {}
	s.mods = nil
	if pairs(s) then
	  for k, v in pairs(s) do
	    self.level.map.attrs(i-1, j-1, k, mods[k] and mods[k](k, v) or v)
	  end
	end
      end

      if define_spot then
	define_spot = table.clone(define_spot)
	assert(define_spot.type, "defineTile auto spot without type field")
	assert(define_spot.subtype, "defineTile auto spot without subtype field")
	define_spot.x = self.data.__import_offset_x+i-1
	define_spot.y = self.data.__import_offset_y+j-1
	-- Questionable Hack(TM):  Allow a mod function to modify the spot
	-- definition on installation.
	local mod = define_spot.mod
	define_spot.mod = nil
	self.spots[#self.spots+1] = mod and mod(define_spot) or define_spot
      end
    end
  end

  if self.add_attrs_later then
    for _, attr in ipairs(self.add_attrs_later) do
      if attr.key == "lite" then
	self.level.map.lites(attr.x, attr.y, true)
      end
      if attr.key == "remember" then
	self.level.map.remembers(attr.x, attr.y, true)
      end
      if attr.key == "special" then
	self.map.room_map[attr.x][attr.y].special = s.special
      end
      if attr.key == "room_map" then
	for k, v in pairs(s.room_map) do
	  self.map.room_map[attr.x][attr.y][k] = v
	end
      end
      self.level.map.attrs(attr.x, attr.y, attr.key, attr.value)
    end
  end

  self:triggerHook{"MapGenerator:Static:subgenRegister", mapfile=self.data.map, list=self.subgen}

  for i = 1, #self.subgen do
    local g = self.subgen[i]
    local data = g.data
    if type(data) == "string" and data == "pass" then data = self.data end

    local map = self.zone.map_class.new(g.w, g.h)
    data.__import_offset_x = self.data.__import_offset_x+g.x
    data.__import_offset_y = self.data.__import_offset_y+g.y
    local generator = require(g.generator).new(self.zone, map, self.level, data)
    local ux, uy, dx, dy, subspots = generator:generate(lev, old_lev)

    if g.overlay then
      self.map:overlay(map, g.x, g.y)
    else
      self.map:import(map, g.x, g.y)
    end

    if not generator.auto_handle_spot_offsets then
      for _, spot in ipairs(subspots) do
	spot.x = spot.x + data.__import_offset_x
	spot.y = spot.y + data.__import_offset_y
      end
    end

    table.append(self.spots, subspots)

    if g.define_up then
      self.gen_map.startx, self.gen_map.starty = ux + self.data.__import_offset_x+g.x, uy + self.data.__import_offset_y+g.y
    end
    if g.define_down then
      self.gen_map.endx, self.gen_map.endy = dx + self.data.__import_offset_x+g.x, dy + self.data.__import_offset_y+g.y
    end
  end

  if self.gen_map.startx and self.gen_map.starty then
    self.map.room_map[self.gen_map.startx][self.gen_map.starty].special = "exit"
  end
  if self.gen_map.endx and self.gen_map.endy then
    self.map.room_map[self.gen_map.endx][self.gen_map.endy].special = "exit"
  end
  return self.gen_map.startx, self.gen_map.starty, self.gen_map.endx, self.gen_map.endy, self.spots
end
