-- $Id: AutomatizerOneOfEditor.lua 3638 2019-07-16 02:36:26Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2019 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Dialog = require 'engine.ui.Dialog'
local ListColumns = require 'engine.ui.ListColumns'
local Button = require 'engine.ui.Button'
local Textbox = require 'engine.ui.Textbox'
local Dropdown = require 'engine.ui.Dropdown'
local Separator = require 'engine.ui.Separator'
local CardLayout = require 'mod.dialogs.ui.CardLayout'
local ShortDropdown = require 'mod.dialogs.ui.ShortDropdown'
local Automatizer = require 'mod.class.Automatizer'

module(..., package.seeall, class.inherit(Dialog))

function _M:init(val_name, group, grp_only, cur_vals, apply_cb)
  self.vals = cur_vals
  self.apply_cb = apply_cb
  local ok_vals = {}
  if group then
    ok_vals = Automatizer.test_allowed_values[group]
    if type(ok_vals) == 'function' then ok_vals = ok_vals() end
    ok_vals = self:buildList(ok_vals)
  end

  Dialog.init(self, val_name, 1, 1)

  self.list = self:buildList(self.vals, true)
  self.c_list = ListColumns.new {
    width = 200,
    nb_items = 8,
    scrollbar = true,
    columns = {
      { name='', width=100, display_prop='name' },
    },
    list = self.list,
    select = function(item, sel) end,
    fct = function() end,
  }
  local vs = Separator.new { dir='vertical', size=250 }

  self.b_ok = Button.new {
    text = 'OK',
    fct = function() self:finish() end,
  }
  self.b_cancel = Button.new {
    text = 'Cancel',
    fct = function() self.key:triggerVirtual('EXIT') end,
  }
  self.b_add = Button.new {
    text = 'Add',
    fct = function() self:addCurVal() end,
  }
  self.b_remove = Button.new {
    text = 'Remove',
    fct = function() self:removeSelected() end,
  }

  local uis = {
    { left=0, top=0, ui=self.c_list },
    { left=self.c_list, top=0, ui=self.b_add },
    { left=self.c_list, top=self.b_add, ui=self.b_remove },
    { hcenter=0, top=self.c_list, ui=vs }
  }
  if group and grp_only then
    self.mode = 'grpsel'
    self.dd_vals = Dropdown.new {
      width = 200,
      fct = function() end,
      list = ok_vals,
      nb_items = math.min(8, #ok_vals),
      scrollbar = (#ok_vals > 8),
    }
    uis[#uis+1] = { left=0, top=vs, ui=self.dd_vals }
  else
    self.tb_val = Textbox.new {
      title = '',
      text = '',
      chars = 20,
      max_len = 99,
      fct = function() end,
    }
    uis[#uis+1] = { left=0, top=vs, ui=self.tb_val }
    if group then
      self.mode = 'text_with_grpsel'
      self.sd_grpsel = ShortDropdown.new {
	text = 'Select',
	list = ok_vals,
	max_shown = 8,
	fct = function(item)
	  if item and item.value and #item.value > 0 then
	    self.tb_val:setText(item.value)
	  end
	end
      }
      uis[#uis+1] = { left=self.tb_val, top=vs, ui=self.sd_grpsel }
    else
      self.mode = 'text'
    end
  end
  table.append(uis, {
    { left=0, bottom=0, ui=self.b_ok },
    { left=self.b_ok, bottom=0, ui=self.b_cancel },
  })
  self:loadUI(uis)
  self:setupUI(true, true)

  self.key:addBinds {
    EXIT = function() game:unregisterDialog(self) end
  }
end

function _M:finish()
  if not self.vals or #self.vals == 0 then
    Dialog:simplePopup('No Values', 'No values selected.')
    return
  end

  if self.apply_cb then self.apply_cb(self.vals) end
  self.key:triggerVirtual('EXIT')
end

function _M:buildList(vals, add_placeholder)
  local function list_cb(_, v) return { name=tostring(v), value=v } end
  local ret = table.maplist(list_cb, vals)
  if #ret == 0 and add_placeholder then
    ret = {
      {
	name = '#{italic}##LIGHT_SLATE#-- No values --#LAST##{normal}#',
	value = '',
      }
    }
  end
  return ret
end

function _M:addCurVal()
  local val
  if self.mode == 'grpsel' then
    local sel = self.dd_vals.c_list.sel or 1
    sel = self.dd_vals.c_list.list[sel]
    val = sel and sel.value
  else
    val = self.tb_val.text
  end

  if not val or #val == 0 then
    Dialog:simplePopup('No Value', 'No value to add.')
    return
  end
  for _, v in ipairs(self.vals) do
    if v == val then
      Dialog:simplePopup('Duplicate Value', 'That value is already on the list.')
      return
    end
  end

  self.vals[#self.vals+1] = val
  self.list = self:buildList(self.vals, true)
  self.c_list:setList(self.list)
end

function _M:removeSelected()
  if not self.vals or #self.vals == 0 then
    Dialog:simplePopup('No Values', 'There is no value to remove.')
    return
  end
  if not self.c_list.sel or self.c_list.sel < 1 or self.c_list.sel > #self.vals then
    Dialog:simplePopup('No Selection', 'No value is selected to remove.')
    return
  end

  table.remove(self.vals, self.c_list.sel)
  self.list = self:buildList(self.vals, true)
  self.c_list:setList(self.list)
end
