-- $Id: AutomatizerRules.lua 3678 2019-09-19 01:25:29Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2019 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Dialog = require 'engine.ui.Dialog'
local List = require 'engine.ui.List'
local Button = require 'engine.ui.Button'
local Checkbox = require 'engine.ui.Checkbox'
local Textzone = require 'engine.ui.Textzone'
local TextzoneList = require 'engine.ui.TextzoneList'
local Textbox = require 'engine.ui.Textbox'
local TweakedTreeList = require 'mod.dialogs.ui.TweakedTreeList'
local Dropdown = require 'engine.ui.Dropdown'
local CardLayout = require 'mod.dialogs.ui.CardLayout'
local ObjectData = require 'mod.class.interface.ObjectData'
local Automatizer = require 'mod.class.Automatizer'
local AutomatizerClauseAction = require 'mod.dialogs.AutomatizerClauseAction'

module(..., package.seeall, class.inherit(Dialog))

local placeholder_name = '#{italic}#[placeholder - click to replace]#{normal}#'

function _M:init()
  Dialog.init(self, 'Automatizer Rules', 800, 600)

  self.working_rules = self:cloneRulesIncoming(Automatizer.atm_rules)

  self.c_active = Checkbox.new {
    title = 'Automatizer is active',
    default = Automatizer.active,
    on_change = function(s) Automatizer.active = s end,
  }
  self.c_action = Dropdown.new {
    width = 120,
    list = {
      { name='Destroy', action='destroy' },
      { name='Pick up', action='pickup' },
      { name='Inscribe', action="inscribe" },
      { name='Configure', action='configure' },
    },
    nb_items = 4,
    fct = function(item) self:selectAction(item) end,
  }
  self.c_global = Checkbox.new {
    title = 'Apply to all characters',
    default = false,
    on_change = function(s)
      if self.cur_rule then
	self.cur_rule.global = s
	self:updateRules()
	self:updateRevertButton()
      end
    end,
  }
  self.c_revert_1 = Button.new {
    text = 'Revert Rule',
    fct = function() self:revertRule() end,
  }

  self:generateRules()
  self.c_rules = List.new {
    width = 150,
    height = self.ih - self.c_active.h - 5,
    scrollbar = true,
    list = self.list,
    fct = function(item, _)
      if item.rule then self:selectRule(item.rule.name) end
    end,
    select = function(_, _) end,
  }

  self.c_rule = TweakedTreeList.new {
    width = 300,
    height = self.ih - math.max(self.c_action.h, self.c_revert_1.h) - self.c_global.h,
    --all_clicks = true,
    scrollbar = true,
    columns = {
      { name='Rule', width=100, display_prop='rule_desc' },
    },
    tree = {},
    fct = function(item, sel, v) self:modifyClause(item) end,
    select = function(item, sel) end,
    on_expand = function(item) end,
    on_drawitem = function(item) end,
  }

  self.c_rule_name = Textbox.new {
    title = ' Name: ',
    text = '',
    chars = 35,
    max_len = 99,
    fct = function() end,
    on_change = function(text) self:fiddleRuleName(text) end,
  }
  self.c_inscription = Textbox.new {
    title = 'Inscription:',
    text = '',
    chars = 30,
    max_len = 99,
    fct = function() end,
    on_change = function(text) self:updateInscription(text) end,
  }

  self.config_cbs = {
    ask_destroy = Checkbox.new {
      title = 'Ask before destroying',
      default = false,
      on_change = function() self:updateConfig() end,
    },
    ask_drop = Checkbox.new {
      title = 'Ask before dropping',
      default = false,
      on_change = function() self:updateConfig() end,
    },
    ask_wear = Checkbox.new {
      title = 'Ask before wearing',
      default = false,
      on_change = function() self:updateConfig() end,
    },
    ask_use = Checkbox.new {
      title = 'Ask before using',
      default = false,
      on_change = function() self:updateConfig() end,
    },
    notify_recharge = Checkbox.new {
      title = 'Notify on recharge',
      default = false,
      on_change = function() self:updateConfig() end,
    },
  }

  self.cards = CardLayout.new {
    destroy = {},
    pickup = {},
    inscribe = {
      { left=0, top=0, ui=self.c_inscription },
    },
    configure = {
      { left=0, top=0, ui=self.config_cbs.ask_destroy },
      { left=self.config_cbs.ask_destroy, top=0, ui=self.config_cbs.ask_drop },
      { left=0, top=self.config_cbs.ask_destroy, ui=self.config_cbs.ask_wear },
      { left=self.config_cbs.ask_wear, top=self.config_cbs.ask_drop, ui=self.config_cbs.ask_use },
      { left=0, top=self.config_cbs.ask_wear, ui=self.config_cbs.notify_recharge },
    },
  }
  self.cards:set_initial_card('destroy')

  self.c_new_rule = Button.new {
    text = "New Rule",
    fct = function() self:createNewRule() end,
  }
  self.c_test_rule = Button.new {
    text = "Test Rule",
    fct = function() self:testRule() end,
  }
  self.c_delete_rule = Button.new {
    text = 'Delete Rule',
    fct = function() self:deleteRule() end,
  }

  self.c_test_clr = Button.new {
    text = "Clear",
    fct = function()
      self:setTestObject(nil)
      self:rebuildRuleTree()
    end,
  }
  self.c_test_clr:generate()	-- to compute width/height
  self.c_test_tgt = TextzoneList.new {
    width = 300 - self.c_test_clr.w,
    height = self.c_test_clr.h,
    weakstore = true,
  }

  self.c_save = Button.new {
    text = "Save",
    fct = function() self:saveRules() end,
  }
  self.c_revert_all = Button.new {
    text = "Revert All",
    fct = function() self:revertRules() end,
  }
  self.c_exit = Button.new {
    text = "Exit",
    fct = function() self:closeDialog() end,
  }

  local desc = Textzone.new {
    width = 330,
    auto_height = true,
    text = [[Select a rule to view or edit its clause tree.  Select a clause in the rule tree to modify or remove it.  You can change a rule's name in the "Name:" field.  "New Rule" creates a new rule; "Test Rule" allows you to check whether the selected rule matches an item in your inventory.]],
  }

  self:loadUI {
    { left=0, top=0, ui=self.c_active },
    { left=0, top=self.c_active.h+5, ui=self.c_rules },
    { left=self.c_rules, top=self.c_active.h+5, ui=self.c_rule_name },
    { left=self.c_rules, top=self.c_rule_name, ui=self.c_test_tgt },
    { left=self.c_test_tgt, top=self.c_rule_name, ui=self.c_test_clr },
    { left=self.c_rules, top=self.c_test_tgt, ui=self.c_new_rule },
    { left=self.c_new_rule, top=self.c_test_tgt, ui=self.c_test_rule },
    { left=self.c_test_rule, top=self.c_test_tgt, ui=self.c_delete_rule },
    { left=self.c_rules, top=self.c_new_rule, ui=self.cards:current_card() },
    { left=self.c_rules, bottom=0, ui=self.c_save },
    { left=self.c_save, bottom=0, ui=self.c_revert_all },
    { left=self.c_revert_all, bottom=0, ui=self.c_exit },
    { left=self.c_rules, bottom=self.c_save, ui=desc },
    { right=0, top=0, ui=self.c_action },
    { right=self.c_action, top=0, ui=self.c_revert_1 },
    { right=0, top=self.c_revert_1, ui=self.c_global },
    { right=0, top=self.c_global, ui=self.c_rule },
  }
  self:setFocus(self.c_rules)
  self:setupUI()

  self.key:addBind('EXIT', function() self:closeDialog() end)

  self:setCurrentRule(nil)
  self:setTestObject(nil)
end

function _M:closeDialog()
  local ch = self:changeCount()
  if not ch then
    game:unregisterDialog(self)
  else
    ch = self:changeDesc(ch)
    local text = 'You have '..ch..'.\n\nDiscard changes and exit dialog?'
    local function cb(ok)
      -- Revert is on 'No' button.
      if not ok then game:unregisterDialog(self) end
    end
    Dialog:yesnoPopup('Automatizer: Exit', text, cb, 'Do not exit', 'Exit', true)
  end
end

function _M:cloneRulesIncoming(rules)
  -- On the incoming pass, we only clone the numeric index, to avoid ugly
  -- hachery around renaming rules, and add an 'orig_name' field for
  -- tracking name changes.
  local ret = {}
  for _, rule in ipairs(rules) do
    local rr = table.clone(rule, true)
    rr.orig_name = rr.name
    ret[#ret+1] = rr
  end
  return ret
end

function _M:cloneRulesOutgoing(rules)
  -- On the outgoing pass, populate both the numeric and by-name indexes
  -- and clean up the 'orig_name' field.  We also clean up leftover
  -- inscription and config{} fields for rules that don't need them.
  local ret = {}
  for _, rule in ipairs(rules) do
    local rr = table.clone(rule, true, {orig_name=true})
    if rr.action ~= 'inscribe' then rr.inscription = nil end
    if rr.action ~= 'configure' then rr.config = nil end
    ret[#ret+1] = rr
    ret[rr.name] = rr
  end
  return ret
end

function _M:updateRevertButton()
  self:toggleDisplay(self.c_revert_1, self.cur_rule and self:ruleChanged(self.cur_rule))
end

function _M:populateCardsFromRule(rule)
  if not rule then return end
  if rule.action == 'inscribe' then
    self.c_inscription:setText(rule.inscription or '')
  elseif rule.action == 'configure' then
    for _, tag in ipairs(ObjectData.config_types) do
      self.config_cbs[tag].checked = rule.config and rule.config[tag] and true or false
    end
  end
end

function _M:setCurrentRule(rule)
  local show = rule and true or false
  local can_delete = rule and rule.auto and true or false
  self.cur_rule = rule
  self:toggleDisplay(self.c_rule_name, show)
  self:toggleDisplay(self.c_test_rule, show)
  self:toggleDisplay(self.c_delete_rule, can_delete)
  self:updateRevertButton()
  if rule then
    ui_util.selDropdownByField(self.c_action, 'action', rule.action)
    self.c_rule_name:setText(rule.name)
    self.c_global.checked = rule.global
    self:populateCardsFromRule(rule)
  end
end

function _M:ruleChanged(rule, ignore_name_change)
  if not rule.orig_name then return false end
  if not ignore_name_change and rule.name ~= rule.orig_name then
    return true
  end
  -- Slight Hack(TM):  We know by construction that the 'orig_name' field
  -- is either absent (for new rules) or the name of an original rule.
  local orig_rule = Automatizer.atm_rules[rule.orig_name]

  -- Hack(TM):  Avoid false/nil comparison weirdness.
  local function diff_bool(a, b)
    return (a and not b) or (b and not a)
  end

  if diff_bool(rule.global, orig_rule.global) then return true end
  if rule.action ~= orig_rule.action then return true end
  if rule.action == 'inscribe' and rule.inscription ~= orig_rule.inscription then
    return true
  end
  if rule.action == 'configure' then
    if not rule.config or not orig_rule.config then return true end
    for _, tag in ipairs(ObjectData.config_types) do
      if diff_bool(rule.config[tag], orig_rule.config[tag]) then
	return true
      end
    end
  end
  return not table.equal(rule.match, orig_rule.match, true)
end

function _M:changeCount()
  local ret, orig = { added=0, changed=0, removed=0 }, {}
  for _, rule in ipairs(self.working_rules) do
    -- Slight Hack(TM):  We know by construction that the 'orig_name' field
    -- is either absent (for new rules) or the name of an original rule.
    if not rule.orig_name then
      ret.added = ret.added + 1
    else
      orig[rule.orig_name] = true
      if self:ruleChanged(rule) then
	ret.changed = ret.changed + 1
      end
    end
  end
  for _, rule in ipairs(Automatizer.atm_rules) do
    if not orig[rule.name] then
      ret.removed = ret.removed + 1
    end
  end

  if ret.added > 0 or ret.changed > 0 or ret.removed > 0 then return ret end
  return nil
end

function _M:updateRules(no_reselect)
  self:generateRules()
  self.c_rules.list = self.list
  self.c_rules:generate()
  if self.cur_rule then
    if self.list_idx_by_name[self.cur_rule.name] then
      self.c_rules:select(self.list_idx_by_name[self.cur_rule.name])
    end
    if not no_reselect then
      self:selectRule(self.cur_rule.name)
    end
  end
end

function _M:changeDesc(ch)
  local ret = {}
  if ch.added and ch.added > 0 then
    ret[#ret+1] = ('%d new rule%s'):format(ch.added, ch.added == 1 and '' or 's')
  end
  if ch.changed and ch.changed > 0 then
    ret[#ret+1] = ('%d changed/renamed rule%s'):format(ch.changed, ch.changed == 1 and '' or 's')
  end
  if ch.removed and ch.removed > 0 then
    ret[#ret+1] = ('%d removed rule%s'):format(ch.removed, ch.removed == 1 and '' or 's')
  end
  return table.concatNice(ret, ', ', ' and ')
end

function _M:saveRules()
  if next(self.collide_names) then
    local text = 'The following rule names are used by multiple rules:\n· ' .. table.concat(table.keys(self.collide_names), '\n· ')
    Dialog:simpleLongPopup('Rule Name Collisions', text, 400)
    return
  end

  local ch = self:changeCount()
  if not ch then
    game:unregisterDialog(self)
    Dialog:simplePopup('Automatizer: Save', 'No changed rules to save')
  else
    ch = self:changeDesc(ch)
    local text = 'Save '..ch..'?'
    local function cb(ok)
      -- 'Save' is on 'No' button
      if not ok then
	local n = 0
	for _, rule in ipairs(self.working_rules) do
	  if rule.auto then n = n + 1 end
	end
	Automatizer.atm_rules = self:cloneRulesOutgoing(self.working_rules)
	Automatizer:writeAutoRules()
	-- Slight Hack(TM):  Now that the rules have been saved, none of
	-- them are new and all of them correspond to an original rule of
	-- the same name.
	for _, rule in ipairs(self.working_rules) do
	  rule.orig_name = rule.name
	end
	self:updateRules()
	Dialog:simplePopup('Automatizer: Save', 'Saved '..n..' automatizer rule(s)')
      end
    end
    Dialog:yesnoPopup('Automatizer: Save', text, cb, 'Cancel', 'Save', true)
  end
end

function _M:revertRule()
  if not self.cur_rule or not self:ruleChanged(self.cur_rule) then return end
  local function cb(ok)
    -- 'Revert' is on the 'No' button.
    if not ok then
      local i = self.rule_idx_by_name[self.cur_rule.name]
      if i then
	if self.cur_rule.orig_name then
	  local rr = table.clone(Automatizer.atm_rules[self.cur_rule.orig_name], true)
	  rr.orig_name = rr.name
	  self.working_rules[i] = rr
	  self:setCurrentRule(rr)
	else
	  table.remove(self.working_rules, i)
	  self:setCurrentRule(nil)
	end
	-- This will rebuild the rule tree for us.
	self:updateRules()
      end
    end
  end
  local title, fmt = 'Revert Rule', 'Revert changes to rule "%s"?'
  if not self.cur_rule.orig_name then
    title, fmt = 'Discard rule', 'Discard unsaved rule "%s"?'
  end
  Dialog:yesnoPopup(title, fmt:format(self.cur_rule.name), cb, 'Do not revert', 'Revert', false, true)
end

function _M:revertRules()
  local ch = self:changeCount()
  if not ch then return end
  ch = self:changeDesc(ch)
  local text = 'You have '..ch..'.\n\nDiscard changes and restore original rules?'
  local function cb(ok)
    -- Revert is on 'No' button.
    if not ok then
      self.working_rules = self:cloneRulesIncoming(Automatizer.atm_rules)
      self:setCurrentRule(nil)
      self:updateRules(true)
      self:rebuildRuleTree()
    end
  end
  Dialog:yesnoPopup('Automatizer: Revert', text, cb, 'Do not revert', 'Revert', true)
end

function _M:updateConfig()
  if self.cur_rule and self.cur_rule.action == 'configure' then
    self.cur_rule.config = self.cur_rule.config or {}
    for _, tag in ipairs(ObjectData.config_types) do
      self.cur_rule.config[tag] = self.config_cbs[tag].checked
    end
    self:updateRevertButton()
    self:updateRules(true)
  end
end

function _M:updateInscription(text)
  if self.cur_rule and self.cur_rule.action == 'inscribe' then
    self.cur_rule.inscription = text
    self:updateRevertButton()
    self:updateRules(true)
  end
end

function _M:selectAction(item)
  if not item or not item.action then return end
  self.cards:show_card(item.action, self)
  if self.cur_rule then
    self.cur_rule.action = item.action
    self:populateCardsFromRule(self.cur_rule)
  end
  self:updateRevertButton()
  self:updateRules(true)
end

function _M:setTestObject(o)
  self.test_obj = o
  if o then
    local text = '#{italic}#Testing vs:#{normal}#  ' .. o:getName()
    self.c_test_tgt:switchItem(o, text)
    self:toggleDisplay(self.c_test_clr, true)
  else
    local text = '#{italic}##GREY#[no test object selected]#LAST##{normal}#'
    self.c_test_tgt:switchItem('@NONE@', text)
    self:toggleDisplay(self.c_test_clr, false)
  end
end

function _M:selectRule(rule_name, o)
  local sel_rule = self.rules_by_name[rule_name]
  if not sel_rule then
    print(('could not find automatizer rule "%s"'):format(tostring(rule)))
    return
  end
  if self.list_idx_by_name[rule_name] then
    self.c_rules:select(self.list_idx_by_name[rule_name])
  end
  self:setCurrentRule(sel_rule)
  self:setTestObject(o)
  self:rebuildRuleTree()
end

function _M:rebuildRuleTree(keep_sel)
  local orig_sel = keep_sel and self.c_rule.sel
  self.rule_tree = {}
  if self.cur_rule then
    self.rule_tree = {
      self:generateRuleTree(self.cur_rule.match, self.cur_rule, 'match')
    }
  end
  self.c_rule.tree = self.rule_tree
  self.c_rule:generate()
  if orig_sel then
    self.c_rule.sel = util.bound(orig_sel, 1, self.c_rule.max)
  end
end

function _M:createNewRule()
  local GetText = require 'engine.dialogs.GetText'
  local function cb(text)
    local action = self.c_action.c_list.list[self.c_action.c_list.sel].action
    local rule = {
      name = text,
      auto = true,
      action = action,
      match = { test='placeholder' },
    }
    if action == 'inscribe' then
      rule.inscription = self.c_inscription.text or ''
    end
    if action == 'configure' then
      rule.config = {}
      for _, tag in ipairs(ObjectData.config_types) do
	rule.config[tag] = self.config_cbs[tag].checked
      end
    end
    table.insert(self.working_rules, rule)
    self:updateRules(true)
    self:selectRule(text)
  end
  local d = GetText.new('New Rule', 'Name: ', 2, 99, cb)
  game:registerDialog(d)
end

function _M:testRule()
  if not self.cur_rule then return end
  local title = 'Test Rule:  ' .. self.cur_rule.name
  local function cb(o, _, _)
    self:selectRule(self.cur_rule.name, o)
    return true
  end
  game.player:showEquipInven(title, nil, cb)
end

function _M:deleteRule(force)
  if not self.cur_rule then return end
  if not force then
    local function cb(ok)
      -- Delet is on 'No' button.
      if not ok then self:deleteRule(true) end
    end
    local text = 'Delete automatizer rule "' .. self.cur_rule.name .. '"?'
    Dialog:yesnoPopup('Delete Automatizer Rule', text, cb, 'Do not delete', 'Delete', true)
    return
  end
  for i, rr in ipairs(self.working_rules) do
    if rr.name == self.cur_rule.name then
      table.remove(self.working_rules, i)
      self:setCurrentRule(nil)
      self:updateRules(true)
      self:rebuildRuleTree()
      break
    end
  end
end

function _M:modifyClause(item)
  if not self.cur_rule or not self.cur_rule.auto then return end
  local function on_change_cb()
    self:rebuildRuleTree(true)
    self:updateRules(true)
    self:updateRevertButton()
  end
  -- We can only do actions on clauses or members of a one-of group.
  if item.subrule or (item.parent and item.parent.test == 'oneof') then
    game:registerDialog(AutomatizerClauseAction.new(self.cur_rule, item.subrule, item.parent, item.child_idx, on_change_cb))
  end
end

function _M:fiddleRuleName(name)
  if not self.cur_rule or self.cur_rule.name == name then return end
  self.cur_rule.name = name
  self:updateRules(true)
  self:updateRevertButton()
end

function _M:generateRules()
  local by_name, idx_by_name, collide = {}, {}, {}
  for i, rule in ipairs(self.working_rules) do
    if by_name[rule.name] then collide[rule.name] = true end
    by_name[rule.name] = rule
    idx_by_name[rule.name] = i
  end

  local auto, manual = {}, {}
  for _, rule in ipairs(self.working_rules) do
    local name = rule.name
    if collide[rule.name] then name = '#LIGHT_RED#'..name..'#LAST#' end
    if rule.global then name = '#{italic}#'..name..'#{normal}#' end
    local pfx = ''
    if not rule.orig_name then
      pfx = pfx .. '!'
    else
      if self:ruleChanged(rule, true) then pfx = pfx .. '*' end
      if rule.name ~= rule.orig_name  then pfx = pfx .. '=' end
    end
    if #pfx > 0 then
      name = '#{bold}##LIGHT_GREEN#'..pfx..'#LAST##{normal}# '..name
    end
    local item = { name=name, rule=rule }
    if rule.auto then auto[#auto+1] = item else manual[#manual+1] = item end
  end

  local list, list_idx_by_name = {}, {}
  if #auto > 0 then
    list[#list+1] = { name='#GREY##{bold}#Generated rules#{normal}##LAST#' }
    table.append(list, auto)
  end
  if #manual > 0 then
    list[#list+1] = { name='#GREY##{bold}#Manual rules#{normal}##LAST#' }
    table.append(list, manual)
  end
  for i, item in ipairs(list) do
    if item.rule then list_idx_by_name[item.rule.name] = i end
  end

  self.list = list
  self.rules_by_name = by_name
  self.rule_idx_by_name = idx_by_name
  self.list_idx_by_name = list_idx_by_name
  self.collide_names = collide
end

local clauses = {
  name = 'item name',
  inscription = 'item inscription',
  type = 'item type',
  subtype = 'item subtype',
  full_type = 'item type/subtype',
  define_as = "item 'define_as' tag",
  discount = 'item discount',
  status = 'item status',
  race = 'player race',
  subrace = 'player subrace',
  class = 'player class',
  god = 'player god',
  descriptor = { 'player @1@ descriptor', 1 },
  level = 'player level',
  skill = { 'player @1@ skill', 1 },
  is_identified = 'item is identified',
  has_ability = { 'player has @1@ ability', 1 },
  has_talent = { 'player has @1@ talent', 1 },
  has_addon = { 'addon "@1@" is enabled', 1 },
}

local function expr_desc(m)
  if type(m) ~= 'table' then return tostring(m) end
  if m.clause and clauses[m.clause] then
    local ret = clauses[m.clause]
    if type(ret) == 'table' then
      local rret = ret[1]
      for i = 1, ret[2] do
	rret = rret:gsub('@'..i..'@', tostring(m[i]))
      end
      ret = rret
    end
    if m.base then
      ret = 'base ' .. ret
    end
    return ret
  end
  return ('??? [clause=%s]'):format(tostring(m.clause))
end

local test_desc = {
  ['=='] = 'equals',
  ['!='] = 'does not equal',
  ['<='] = 'is less than or equal to',
  ['>='] = 'is greater than or equal to',
  ['<'] = 'is less than',
  ['>'] = 'is greater than',
  contains = 'contains',
  matches = 'matches',
}

function _M:generateRuleTree(m, parent, c_idx)
  local item = { subrule=m, parent=parent, child_idx=c_idx }

  local function r_desc(desc, alt_test)
    if not self.test_obj then return tstring { desc } end
    local ok = Automatizer:itemMatches(alt_test or m, self.test_obj)
    return tstring { {'color', (ok and 'LIGHT_GREEN' or 'LIGHT_RED')}, desc, {'color', 'LAST'} }
  end

  if m.test == 'placeholder' then
    item.rule_desc = placeholder_name:toTString()
    return item
  end
  if m.test == 'and' or m.test == 'or' then
    item.rule_desc = r_desc(m.test == 'and' and 'All of the following are true:' or 'At least one of the following is true:')
    local nodes = {}
    for i, sm in ipairs(m) do
      nodes[#nodes+1] = self:generateRuleTree(sm, m, i)
    end
    if #nodes == 0 then
      nodes = { { rule_desc=placeholder_name:toTString() } }
    end
    item.nodes = nodes
    item.shown = true
    return item
  end
  if m.test == 'not' then
    item.rule_desc = r_desc('The following is not true:')
    item.nodes = { self:generateRuleTree(m[1], m, 1) }
    item.shown = true
    return item
  end
  if m.test == 'oneof' then
    item.rule_desc = r_desc(('%s is one of the following:'):format(expr_desc(m[1]):capitalize()))
    local nodes = {}
    for i = 2, #m do
      nodes[#nodes+1] = { rule_desc=r_desc(expr_desc(m[i]), {test='==', m[1], m[i]}), parent=m, child_idx=i }
    end
    item.nodes = nodes
    item.shown = true
    return item
  end
  if m.test == 'inrange' then
    item.rule_desc = r_desc(('%s is between %s and %s (inclusive)'):format(expr_desc(m[1]):capitalize(), expr_desc(m[2]), expr_desc(m[3])))
    return item
  end
  if m.test and test_desc[m.test] then
    item.rule_desc = r_desc(('%s %s %s'):format(expr_desc(m[1]):capitalize(), test_desc[m.test], expr_desc(m[2])):capitalize())
    return item
  end
  if m.clause then
    item.rule_desc = r_desc(expr_desc(m):capitalize())
    return item
  end
  if m.clause == 'inventory' or m.clause == 'equipment' then
    item.rule_desc = r_desc(m.clause == 'inventory' and 'Something in your inventory matches:' or 'Something in your equipment matches:')
    local nodes = {}
    for i, sm in ipairs(m) do
      nodes[#nodes+1] = self:generateRuleTree(sm, m, i)
    end
    if #nodes == 0 then
      nodes = { { rule_desc=placeholder_name:toTString() } }
    end
    item.nodes = nodes
    item.shown = true
    return item
  end

  item.rule_desc = r_desc(('??? [test=%s, clause=%s]'):format(tostring(m.test), tostring(m.clause)))
  return item
end
