-- $Id: ShowInvenFloor.lua 3893 2020-04-21 02:31:51Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2020 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Dialog = require "engine.ui.Dialog"
local ListColumns = require "engine.ui.ListColumns"
local TextzoneList = require "engine.ui.TextzoneList"
local Separator = require "engine.ui.Separator"
local ShowInventory = require 'engine.dialogs.ShowInventory'

module(..., package.seeall, class.inherit(ShowInventory))

-- We duplicate ShowInventory:init() in place with minor tweaks, because we
-- need to change the list column headers before our custom :generateList()
-- gets called.
function _M:init(title, inven, filter, action, actor)
  self.inven = inven
  self.filter = filter
  self.action = action
  self.actor = actor
  Dialog.init(self, title or "Inventory", math.max(800, game.w * 0.8), math.max(600, game.h * 0.8))

  self.c_desc = TextzoneList.new {
    width = math.floor(self.iw / 2 - 10),
    height = self.ih,
    no_color_bleed = true,
  }

  self.c_list = ListColumns.new{
    width = math.floor(self.iw / 2 - 10),
    height = self.ih - 10,
    sortable = true,
    scrollbar = true,
    columns = {
      -- Tweaks:  widen the 'char' column to stop the @#$%ing annoying
      -- wiggling, and allow the 'object' column's direct_draw() method to
      -- operate in the absence of an 'object' column value.
      {name="", width={26,"fixed"}, display_prop="char", sort="id"},
      {name="", width={24,"fixed"}, display_prop="object", sort="sortname", direct_draw=function(item, x, y) if item.object then item.object:toScreen(nil, x+4, y, 16, 16) end end},
      {name="Inventory", width=72, display_prop="name", sort="sortname"},
      {name="Category", width=20, display_prop="cat", sort="cat"},
      {name="Enc.", width=8, display_prop="encumberance", sort="encumberance"},
    },
    list = {},
    fct = function(item) self:use(item) end,
    select = function(item, sel) self:select(item) end
  }

  self:generateList()

  self:loadUI {
    {left=0, top=0, ui=self.c_list},
    {right=0, top=0, ui=self.c_desc},
    {hcenter=0, top=5, ui=Separator.new{dir="horizontal", size=self.ih - 10}},
  }
  self:setFocus(self.c_list)
  self:setupUI()

  self.key:addCommands{
    __TEXTINPUT = function(c)
      if self.list and self.list.chars[c] then
	self:use(self.list[self.list.chars[c]])
      end
    end,
  }
  self.key:addBinds{
    EXIT = function() game:unregisterDialog(self) end,
  }
end

function _M:use(item)
  local dont_end = false
  if item and item.object then
    dont_end = self.action(item.object, item.item, item.floor)
  end
  self:generateList()
  self:select(self.c_list.list[self.c_list.sel])
  if not dont_end then game:unregisterDialog(self) end
end

-- Replaced to add floor items.
function _M:generateList()
  local x, y = self.actor.x, self.actor.y
  local list = { chars={} }

  local do_inv, do_floor = false, false
  if self.filter then
    for _, o in ipairs(self.inven or {}) do
      if self.filter(o) then
	do_inv = true
	break
      end
    end
    if x and y then
      for i = 1, game.level.map:getObjectTotal(x, y) do
	if self.filter(game.level.map:getObject(x, y, i)) then
	  do_floor = true
	  break
	end
      end
    end
  else
    do_inv = self.inven and #self.inven > 0
    do_floor = x and y and game.level.map:getObjectTotal(x, y) > 0
  end

  local ch = 1
  if do_inv then
    list[#list+1] = {
      id = #list + 1,
      char = '',
      name = ('#{bold}#In inventory#{normal}#'):toTString(),
      sortname = '1:',
      color = {0x90, 0x90, 0x90},
      cat = '',
      sort_cat = '1:',
      encumberance = '',
      sort_enc = '1:',
      desc = 'Items in your inventory',
    }
    for item, o in ipairs(self.inven or {}) do
      if not self.filter or self.filter(o) then
	local char = self:makeKeyChar(ch)

	local enc = 0
	o:forAllStack(function(o) enc=enc+o.encumber end)

	list[#list+1] = {
	  id = #list + 1,
	  char = char,
	  name = o:getName(),
	  sortname = '1:' .. o:getName():toString():removeColorCodes(),
	  color = o:getDisplayColor(),
	  object = o,
	  item = item,
	  cat = o.subtype,
	  sort_cat = '1:' .. (o.subtype or ''),
	  encumberance = enc,
	  sort_enc = ('1:%d'):format(enc),
	  desc = o:getDesc(),
	}
	list.chars[char] = #list
	ch = ch + 1
      end
    end
  end

  if do_floor then
    list[#list+1] = {
      id = #list + 1,
      char = '',
      name = ('#{bold}#On floor#{normal}#'):toTString(),
      sortname = '2:',
      color = {0x90, 0x90, 0x90},
      cat = '',
      sort_cat = '2:',
      encumberance = '',
      sort_enc = '2:',
      desc = 'Items on the floor at your feet',
    }
    for i = 1, game.level.map:getObjectTotal(x, y) do
      local o = game.level.map:getObject(x, y, i)
      if not self.filter or self.filter(o) then
	local char = self:makeKeyChar(ch)

	local enc = 0
	o:forAllStack(function(o) enc=enc+o.encumber end)

	list[#list+1] = {
	  id = #list + 1,
	  char = char,
	  name = o:getName(),
	  sortname = '2:' .. o:getName():toString():removeColorCodes(),
	  color = o:getDisplayColor(),
	  object = o,
	  item = i,
	  floor = true,
	  cat = o.subtype,
	  sort_cat = '2:' .. (o.subtype or ''),
	  encumberance = enc,
	  sort_enc = ('2:%d'):format(enc),
	  desc = o:getDesc(),
	}
	list.chars[char] = #list
	ch = ch + 1
      end
    end
  end

  self.list = list
  self.c_list:setList(list)

  -- Slight Hack(TM):  If the generated inventory/floor list is empty, it
  -- won't get properly pushed into the corresponding ListColumns list, and
  -- the 'chars' table we generated will be missing, causing an error when
  -- the user types anything.
  if #self.list == 0 then self.c_list.list.chars = {} end
end

function _M:inventoryLiveUpdate()
  self:generateList()
  self:select(self.c_list.list[self.c_list.sel])
end
