-- $Id: ShortDropdown.lua 3633 2019-07-14 04:09:31Z dsb $
-- TE4 - T-Engine 4
-- Copyright (C) 2009 - 2018 Nicolas Casalini
-- Copyright © 2019 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Nicolas Casalini "DarkGod"
-- darkgod@te4.org

-- A modified version of engine.ui.Dropdown that displays a specified text
-- instead of the selected dropdown list item.  Typically the display text
-- will be shorter than the list width (hence the name), and separate
-- widths are maintained for the two.  The intended use case is a dropdown
-- that is meant to trigger an action rather than select an item.
require "engine.class"
local Base = require "engine.ui.Base"
local Focusable = require "engine.ui.Focusable"
local List = require "engine.ui.List"
local Dialog = require "engine.ui.Dialog"

module(..., package.seeall, class.inherit(Base, Focusable))

function _M:init(t)
  self.text = assert(t.text, 'no dropdown text')
  self.fct = assert(t.fct, "no dropdown fct")
  self.list = assert(t.list, "no dropdown list")
  self.max_shown = t.max_shown
  self.text_color = t.text_color or colors_simple.WHITE
  self.on_select = t.on_select
  self.display_prop = t.display_prop or "name"
  self.alpha = self.alpha or 0.7
  self.scrollbar = t.max_shown and t.max_shown < #self.list

  Base.init(self, t)
end

function _M:generate()
  self.mouse:reset()
  self.key:reset()

  -- Draw UI
  self.w = self.font:size(self.text) + 18
  self.h = self.font_h + 9
  self.height = self.h

  self.frame = self:makeFrame("ui/textbox", self.w, self.h)
  self.frame_sel = self:makeFrame("ui/textbox-sel", self.w, self.h)

  -- Add UI controls
  self.mouse:registerZone(0, 0, self.w, self.h, function(button, x, y, xrel, yrel, bx, by, event)
    if event == "button" and button == "left" then self:showSelect() end
  end)
  self.key:addBind("ACCEPT", function() self:showSelect() end)
end

function _M:positioned(x, y, sx, sy)
  -- Compute list width as needed.
  if not self.list_w then
    self.list_w = 0
    for _, item in ipairs(self.list) do
      self.list_w = math.max(self.list_w, self.font:size(item.name) + 12)
    end
  end
  if self.scrollbar then self.list_w = self.list_w + 20 end
  local nb_items = self.max_shown and math.min(self.max_shown, #self.list) or #self.list

  self.c_list = List.new{width=self.list_w, list=self.list, select=self.on_select, display_prop=self.display_prop, scrollbar=self.scrollbar, nb_items=nb_items, fct=function()
    game:unregisterDialog(self.popup)
    self:sound("button")
    self.fct(self.c_list.list[self.c_list.sel])
  end}
  self.popup = Dialog.new(nil, self.list_w, self.c_list.h, sx, sy + self.h, nil, nil, false, "simple")
  self.popup.frame.a = self.alpha
  self.popup:loadUI{{left=0, top=0, ui=self.c_list}}
  self.popup:setupUI(true, true)
  self.popup.key:addBind("EXIT", function()
    game:unregisterDialog(self.popup)
    self.c_list.sel = self.previous
    self:sound("button")
    self.fct(self.c_list.list[self.c_list.sel])
  end)
end

function _M:showSelect()
  self.previous = self.c_list.sel
  game:registerDialog(self.popup)
end

function _M:display(x, y, nb_keyframes)
  if self.focused then
    self:drawFrame(self.frame_sel, x, y)
  else
    self:drawFrame(self.frame, x, y)
    if self.focus_decay then
      self:drawFrame(self.frame_sel, x, y, 1, 1, 1, self.focus_decay / self.focus_decay_max_d)
      self.focus_decay = self.focus_decay - nb_keyframes
      if self.focus_decay <= 0 then self.focus_decay = nil end
    end
  end

  if not self._tex then
    self._tex = self:drawFontLine(self.font, self.text, self.w, self.text_color[1], self.text_color[2], self.text_color[3])
  end
  local cy = (self.c_list.fh - self.c_list.font_h) / 2
  if self.text_shadow then
    self:textureToScreen(self._tex, x + 1 + self.frame_sel.b4.w, y + 1 + cy, 0, 0, 0, self.text_shadow)
  end
  self:textureToScreen(self._tex, x + self.frame_sel.b4.w, y + cy, 1, 1, 1, 1)
end
